@@ -1,31 +1,34 @@
-use crate::builder::{ConfigBuilder, WantsCipherSuites};
+use crate::builder::ConfigBuilder;
 use crate::common_state::{CommonState, Protocol, Side};
 use crate::conn::{ConnectionCommon, ConnectionCore};
+use crate::crypto::{CryptoProvider, SupportedKxGroup};
 use crate::enums::{CipherSuite, ProtocolVersion, SignatureScheme};
 use crate::error::Error;
-use crate::kx::SupportedKxGroup;
 #[cfg(feature = "logging")]
 use crate::log::trace;
 use crate::msgs::enums::NamedGroup;
 use crate::msgs::handshake::ClientExtension;
 use crate::msgs::persist;
 use crate::sign;
-use crate::suites::SupportedCipherSuite;
-use crate::verify;
+use crate::suites::{ExtractedSecrets, SupportedCipherSuite};
 use crate::versions;
-#[cfg(feature = "secret_extraction")]
-use crate::ExtractedSecrets;
 use crate::KeyLog;
+#[cfg(feature = "ring")]
+use crate::WantsVerifier;
+use crate::{verify, WantsVersions};
 
 use super::handy::{ClientSessionMemoryCache, NoClientSessionStorage};
 use super::hs;
 
-use std::error::Error as StdError;
-use std::marker::PhantomData;
-use std::net::IpAddr;
-use std::ops::{Deref, DerefMut};
-use std::sync::Arc;
-use std::{fmt, io, mem};
+use pki_types::ServerName;
+
+use alloc::sync::Arc;
+use alloc::vec::Vec;
+use core::fmt;
+use core::marker::PhantomData;
+use core::mem;
+use core::ops::{Deref, DerefMut};
+use std::io;
 
 /// A trait for the ability to store client session data, so that sessions
 /// can be resumed in future connections.
@@ -37,9 +40,9 @@ use std::{fmt, io, mem};
 /// `set_`, `insert_`, `remove_` and `take_` operations are mutating; this isn't
 /// expressed in the type system to allow implementations freedom in
 /// how to achieve interior mutability.  `Mutex` is a common choice.
-pub trait ClientSessionStore: Send + Sync {
+pub trait ClientSessionStore: fmt::Debug + Send + Sync {
     /// Remember what `NamedGroup` the given server chose.
-    fn set_kx_hint(&self, server_name: &ServerName, group: NamedGroup);
+    fn set_kx_hint(&self, server_name: ServerName<'static>, group: NamedGroup);
 
     /// This should return the value most recently passed to `set_kx_hint`
     /// for the given `server_name`.
@@ -47,18 +50,25 @@ pub trait ClientSessionStore: Send + Sync {
     /// If `None` is returned, the caller chooses the first configured group,
     /// and an extra round trip might happen if that choice is unsatisfactory
     /// to the server.
-    fn kx_hint(&self, server_name: &ServerName) -> Option<NamedGroup>;
+    fn kx_hint(&self, server_name: &ServerName<'_>) -> Option<NamedGroup>;
 
     /// Remember a TLS1.2 session.
     ///
     /// At most one of these can be remembered at a time, per `server_name`.
-    fn set_tls12_session(&self, server_name: &ServerName, value: persist::Tls12ClientSessionValue);
+    fn set_tls12_session(
+        &self,
+        server_name: ServerName<'static>,
+        value: persist::Tls12ClientSessionValue,
+    );
 
     /// Get the most recently saved TLS1.2 session for `server_name` provided to `set_tls12_session`.
-    fn tls12_session(&self, server_name: &ServerName) -> Option<persist::Tls12ClientSessionValue>;
+    fn tls12_session(
+        &self,
+        server_name: &ServerName<'_>,
+    ) -> Option<persist::Tls12ClientSessionValue>;
 
     /// Remove and forget any saved TLS1.2 session for `server_name`.
-    fn remove_tls12_session(&self, server_name: &ServerName);
+    fn remove_tls12_session(&self, server_name: &ServerName<'static>);
 
     /// Remember a TLS1.3 ticket that might be retrieved later from `take_tls13_ticket`, allowing
     /// resumption of this session.
@@ -69,7 +79,7 @@ pub trait ClientSessionStore: Send + Sync {
     /// simultaneously.
     fn insert_tls13_ticket(
         &self,
-        server_name: &ServerName,
+        server_name: ServerName<'static>,
         value: persist::Tls13ClientSessionValue,
     );
 
@@ -78,27 +88,35 @@ pub trait ClientSessionStore: Send + Sync {
     /// Implementations of this trait must return each value provided to `add_tls13_ticket` _at most once_.
     fn take_tls13_ticket(
         &self,
-        server_name: &ServerName,
+        server_name: &ServerName<'static>,
     ) -> Option<persist::Tls13ClientSessionValue>;
 }
 
 /// A trait for the ability to choose a certificate chain and
 /// private key for the purposes of client authentication.
-pub trait ResolvesClientCert: Send + Sync {
-    /// With the server-supplied acceptable issuers in `acceptable_issuers`,
-    /// the server's supported signature schemes in `sigschemes`,
-    /// return a certificate chain and signing key to authenticate.
+pub trait ResolvesClientCert: fmt::Debug + Send + Sync {
+    /// Resolve a client certificate chain/private key to use as the client's
+    /// identity.
     ///
-    /// `acceptable_issuers` is undecoded and unverified by the rustls
-    /// library, but it should be expected to contain a DER encodings
-    /// of X501 NAMEs.
+    /// `root_hint_subjects` is an optional list of certificate authority
+    /// subject distinguished names that the client can use to help
+    /// decide on a client certificate the server is likely to accept. If
+    /// the list is empty, the client should send whatever certificate it
+    /// has. The hints are expected to be DER-encoded X.500 distinguished names,
+    /// per [RFC 5280 A.1]. See [`crate::DistinguishedName`] for more information
+    /// on decoding with external crates like `x509-parser`.
     ///
-    /// Return None to continue the handshake without any client
+    /// `sigschemes` is the list of the [`SignatureScheme`]s the server
+    /// supports.
+    ///
+    /// Return `None` to continue the handshake without any client
     /// authentication.  The server may reject the handshake later
     /// if it requires authentication.
+    ///
+    /// [RFC 5280 A.1]: https://www.rfc-editor.org/rfc/rfc5280#appendix-A.1
     fn resolve(
         &self,
-        acceptable_issuers: &[&[u8]],
+        root_hint_subjects: &[&[u8]],
         sigschemes: &[SignatureScheme],
     ) -> Option<Arc<sign::CertifiedKey>>;
 
@@ -106,32 +124,29 @@ pub trait ResolvesClientCert: Send + Sync {
     fn has_certs(&self) -> bool;
 }
 
-/// Common configuration for (typically) all connections made by
-/// a program.
+/// Common configuration for (typically) all connections made by a program.
 ///
-/// Making one of these can be expensive, and should be
-/// once per process rather than once per connection.
+/// Making one of these is cheap, though one of the inputs may be expensive: gathering trust roots
+/// from the operating system to add to the [`RootCertStore`] passed to `with_root_certificates()`
+/// (the rustls-native-certs crate is often used for this) may take on the order of a few hundred
+/// milliseconds.
 ///
-/// These must be created via the [`ClientConfig::builder()`] function.
+/// These must be created via the [`ClientConfig::builder()`] or [`ClientConfig::builder_with_provider()`]
+/// function.
 ///
 /// # Defaults
 ///
-/// * [`ClientConfig::max_fragment_size`]: the default is `None`: TLS packets are not fragmented to a specific size.
+/// * [`ClientConfig::max_fragment_size`]: the default is `None` (meaning 16kB).
 /// * [`ClientConfig::resumption`]: supports resumption with up to 256 server names, using session
 ///    ids or tickets, with a max of eight tickets per server.
 /// * [`ClientConfig::alpn_protocols`]: the default is empty -- no ALPN protocol is negotiated.
 /// * [`ClientConfig::key_log`]: key material is not logged.
-#[derive(Clone)]
+///
+/// [`RootCertStore`]: crate::RootCertStore
+#[derive(Debug)]
 pub struct ClientConfig {
-    /// List of ciphersuites, in preference order.
-    pub(super) cipher_suites: Vec<SupportedCipherSuite>,
-
-    /// List of supported key exchange algorithms, in preference order -- the
-    /// first element is the highest priority.
-    ///
-    /// The first element in this list is the _default key share algorithm_,
-    /// and in TLS1.3 a key share for it is sent in the client hello.
-    pub(super) kx_groups: Vec<&'static SupportedKxGroup>,
+    /// Source of randomness and other crypto.
+    pub(super) provider: Arc<CryptoProvider>,
 
     /// Which ALPN protocols we include in our client hello.
     /// If empty, no ALPN extension is sent.
@@ -140,13 +155,17 @@ pub struct ClientConfig {
     /// How and when the client can resume a previous session.
     pub resumption: Resumption,
 
-    /// The maximum size of TLS message we'll emit.  If None, we don't limit TLS
-    /// message lengths except to the 2**16 limit specified in the standard.
+    /// The maximum size of plaintext input to be emitted in a single TLS record.
+    /// A value of None is equivalent to the [TLS maximum] of 16 kB.
     ///
     /// rustls enforces an arbitrary minimum of 32 bytes for this field.
-    /// Out of range values are reported as errors from ClientConnection::new.
+    /// Out of range values are reported as errors from [ClientConnection::new].
     ///
-    /// Setting this value to the TCP MSS may improve latency for stream-y workloads.
+    /// Setting this value to a little less than the TCP MSS may improve latency
+    /// for stream-y workloads.
+    ///
+    /// [TLS maximum]: https://datatracker.ietf.org/doc/html/rfc8446#section-5.1
+    /// [ClientConnection::new]: crate::client::ClientConnection::new
     pub max_fragment_size: Option<usize>,
 
     /// How to decide what client auth certificate/keys to use.
@@ -171,8 +190,6 @@ pub struct ClientConfig {
 
     /// Allows traffic secrets to be extracted after the handshake,
     /// e.g. for kTLS setup.
-    #[cfg(feature = "secret_extraction")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "secret_extraction")))]
     pub enable_secret_extraction: bool,
 
     /// Whether to send data on the first flight ("early data") in
@@ -198,25 +215,68 @@ pub enum Tls12Resumption {
     SessionIdOrTickets,
 }
 
-impl fmt::Debug for ClientConfig {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        f.debug_struct("ClientConfig")
-            .field("alpn_protocols", &self.alpn_protocols)
-            .field("resumption", &self.resumption)
-            .field("max_fragment_size", &self.max_fragment_size)
-            .field("enable_sni", &self.enable_sni)
-            .field("enable_early_data", &self.enable_early_data)
-            .finish_non_exhaustive()
+impl Clone for ClientConfig {
+    fn clone(&self) -> Self {
+        Self {
+            provider: Arc::<CryptoProvider>::clone(&self.provider),
+            resumption: self.resumption.clone(),
+            alpn_protocols: self.alpn_protocols.clone(),
+            max_fragment_size: self.max_fragment_size,
+            client_auth_cert_resolver: Arc::clone(&self.client_auth_cert_resolver),
+            versions: self.versions,
+            enable_sni: self.enable_sni,
+            verifier: Arc::clone(&self.verifier),
+            key_log: Arc::clone(&self.key_log),
+            enable_secret_extraction: self.enable_secret_extraction,
+            enable_early_data: self.enable_early_data,
+        }
     }
 }
 
 impl ClientConfig {
-    /// Create a builder to build up the client configuration.
+    /// Create a builder for a client configuration with the default
+    /// [`CryptoProvider`]: [`crate::crypto::ring::default_provider`] and safe ciphersuite and
+    /// protocol defaults.
     ///
     /// For more information, see the [`ConfigBuilder`] documentation.
-    pub fn builder() -> ConfigBuilder<Self, WantsCipherSuites> {
+    #[cfg(feature = "ring")]
+    pub fn builder() -> ConfigBuilder<Self, WantsVerifier> {
+        // Safety: we know the *ring* provider's ciphersuites are compatible with the safe default protocol versions.
+        Self::builder_with_provider(crate::crypto::ring::default_provider().into())
+            .with_safe_default_protocol_versions()
+            .unwrap()
+    }
+
+    /// Create a builder for a client configuration with the default
+    /// [`CryptoProvider`]: [`crate::crypto::ring::default_provider`], safe ciphersuite defaults and
+    /// the provided protocol versions.
+    ///
+    /// Panics if provided an empty slice of supported versions.
+    ///
+    /// For more information, see the [`ConfigBuilder`] documentation.
+    #[cfg(feature = "ring")]
+    pub fn builder_with_protocol_versions(
+        versions: &[&'static versions::SupportedProtocolVersion],
+    ) -> ConfigBuilder<Self, WantsVerifier> {
+        // Safety: we know the *ring* provider's ciphersuites are compatible with all protocol version choices.
+        Self::builder_with_provider(crate::crypto::ring::default_provider().into())
+            .with_protocol_versions(versions)
+            .unwrap()
+    }
+
+    /// Create a builder for a client configuration with a specific [`CryptoProvider`].
+    ///
+    /// This will use the provider's configured ciphersuites. You must additionally choose
+    /// which protocol versions to enable, using `with_protocol_versions` or
+    /// `with_safe_default_protocol_versions` and handling the `Result` in case a protocol
+    /// version is not supported by the provider's ciphersuites.
+    ///
+    /// For more information, see the [`ConfigBuilder`] documentation.
+    pub fn builder_with_provider(
+        provider: Arc<CryptoProvider>,
+    ) -> ConfigBuilder<Self, WantsVersions> {
         ConfigBuilder {
-            state: WantsCipherSuites(()),
+            state: WantsVersions { provider },
             side: PhantomData,
         }
     }
@@ -227,28 +287,44 @@ impl ClientConfig {
     pub(crate) fn supports_version(&self, v: ProtocolVersion) -> bool {
         self.versions.contains(v)
             && self
+                .provider
                 .cipher_suites
                 .iter()
                 .any(|cs| cs.version().version == v)
     }
 
+    pub(crate) fn supports_protocol(&self, proto: Protocol) -> bool {
+        self.provider
+            .cipher_suites
+            .iter()
+            .any(|cs| cs.usable_for_protocol(proto))
+    }
+
     /// Access configuration options whose use is dangerous and requires
     /// extra care.
-    #[cfg(feature = "dangerous_configuration")]
-    pub fn dangerous(&mut self) -> danger::DangerousClientConfig {
+    pub fn dangerous(&mut self) -> danger::DangerousClientConfig<'_> {
         danger::DangerousClientConfig { cfg: self }
     }
 
     pub(super) fn find_cipher_suite(&self, suite: CipherSuite) -> Option<SupportedCipherSuite> {
-        self.cipher_suites
+        self.provider
+            .cipher_suites
             .iter()
             .copied()
             .find(|&scs| scs.suite() == suite)
     }
+
+    pub(super) fn find_kx_group(&self, group: NamedGroup) -> Option<&'static dyn SupportedKxGroup> {
+        self.provider
+            .kx_groups
+            .iter()
+            .copied()
+            .find(|skxg| skxg.name() == group)
+    }
 }
 
 /// Configuration for how/when a client is allowed to resume a previous session.
-#[derive(Clone)]
+#[derive(Clone, Debug)]
 pub struct Resumption {
     /// How we store session data or tickets. The default is to use an in-memory
     /// [ClientSessionMemoryCache].
@@ -297,14 +373,6 @@ impl Resumption {
     }
 }
 
-impl fmt::Debug for Resumption {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        f.debug_struct("Resumption")
-            .field("tls12_resumption", &self.tls12_resumption)
-            .finish()
-    }
-}
-
 impl Default for Resumption {
     /// Create an in-memory session store resumption with up to 256 server names, allowing
     /// a TLS 1.2 session to resume with a session id or RFC 5077 ticket.
@@ -313,90 +381,15 @@ impl Default for Resumption {
     }
 }
 
-/// Encodes ways a client can know the expected name of the server.
-///
-/// This currently covers knowing the DNS name of the server, but
-/// will be extended in the future to supporting privacy-preserving names
-/// for the server ("ECH").  For this reason this enum is `non_exhaustive`.
-///
-/// # Making one
-///
-/// If you have a DNS name as a `&str`, this type implements `TryFrom<&str>`,
-/// so you can do:
-///
-/// ```
-/// # use rustls::ServerName;
-/// ServerName::try_from("example.com").expect("invalid DNS name");
-///
-/// // or, alternatively...
-///
-/// let x = "example.com".try_into().expect("invalid DNS name");
-/// # let _: ServerName = x;
-/// ```
-#[non_exhaustive]
-#[derive(Clone, Debug, Eq, Hash, PartialEq)]
-pub enum ServerName {
-    /// The server is identified by a DNS name.  The name
-    /// is sent in the TLS Server Name Indication (SNI)
-    /// extension.
-    DnsName(verify::DnsName),
-
-    /// The server is identified by an IP address. SNI is not
-    /// done.
-    IpAddress(IpAddr),
-}
-
-impl ServerName {
-    /// Return the name that should go in the SNI extension.
-    /// If [`None`] is returned, the SNI extension is not included
-    /// in the handshake.
-    pub(crate) fn for_sni(&self) -> Option<webpki::DnsNameRef> {
-        match self {
-            Self::DnsName(dns_name) => Some(dns_name.0.as_ref()),
-            Self::IpAddress(_) => None,
-        }
-    }
-}
-
-/// Attempt to make a ServerName from a string by parsing
-/// it as a DNS name.
-impl TryFrom<&str> for ServerName {
-    type Error = InvalidDnsNameError;
-    fn try_from(s: &str) -> Result<Self, Self::Error> {
-        match webpki::DnsNameRef::try_from_ascii_str(s) {
-            Ok(dns) => Ok(Self::DnsName(verify::DnsName(dns.into()))),
-            Err(webpki::InvalidDnsNameError) => match s.parse() {
-                Ok(ip) => Ok(Self::IpAddress(ip)),
-                Err(_) => Err(InvalidDnsNameError),
-            },
-        }
-    }
-}
-
-/// The provided input could not be parsed because
-/// it is not a syntactically-valid DNS Name.
-#[derive(Debug)]
-pub struct InvalidDnsNameError;
-
-impl fmt::Display for InvalidDnsNameError {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        f.write_str("invalid dns name")
-    }
-}
-
-impl StdError for InvalidDnsNameError {}
-
 /// Container for unsafe APIs
-#[cfg(feature = "dangerous_configuration")]
 pub(super) mod danger {
-    use std::sync::Arc;
+    use alloc::sync::Arc;
 
     use super::verify::ServerCertVerifier;
     use super::ClientConfig;
 
     /// Accessor for dangerous configuration options.
     #[derive(Debug)]
-    #[cfg_attr(docsrs, doc(cfg(feature = "dangerous_configuration")))]
     pub struct DangerousClientConfig<'a> {
         /// The underlying ClientConfig
         pub cfg: &'a mut ClientConfig,
@@ -419,6 +412,7 @@ enum EarlyDataState {
     Rejected,
 }
 
+#[derive(Debug)]
 pub(super) struct EarlyData {
     state: EarlyDataState,
     left: usize,
@@ -540,7 +534,7 @@ impl ClientConnection {
     /// Make a new ClientConnection.  `config` controls how
     /// we behave in the TLS protocol, `name` is the
     /// name of the server we want to talk to.
-    pub fn new(config: Arc<ClientConfig>, name: ServerName) -> Result<Self, Error> {
+    pub fn new(config: Arc<ClientConfig>, name: ServerName<'static>) -> Result<Self, Error> {
         Ok(Self {
             inner: ConnectionCore::for_client(config, name, Vec::new(), Protocol::Tcp)?.into(),
         })
@@ -587,6 +581,12 @@ impl ClientConnection {
         self.inner.core.is_early_data_accepted()
     }
 
+    /// Extract secrets, so they can be used when configuring kTLS, for example.
+    /// Should be used with care as it exposes secret key material.
+    pub fn dangerous_extract_secrets(self) -> Result<ExtractedSecrets, Error> {
+        self.inner.dangerous_extract_secrets()
+    }
+
     fn write_early_data(&mut self, data: &[u8]) -> io::Result<usize> {
         self.inner
             .core
@@ -598,13 +598,6 @@ impl ClientConnection {
                     .send_early_plaintext(&data[..sz])
             })
     }
-
-    /// Extract secrets, so they can be used when configuring kTLS, for example.
-    #[cfg(feature = "secret_extraction")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "secret_extraction")))]
-    pub fn extract_secrets(self) -> Result<ExtractedSecrets, Error> {
-        self.inner.extract_secrets()
-    }
 }
 
 impl Deref for ClientConnection {
@@ -643,17 +636,14 @@ impl From<ClientConnection> for crate::Connection {
 impl ConnectionCore<ClientConnectionData> {
     pub(crate) fn for_client(
         config: Arc<ClientConfig>,
-        name: ServerName,
+        name: ServerName<'static>,
         extra_exts: Vec<ClientExtension>,
         proto: Protocol,
     ) -> Result<Self, Error> {
         let mut common_state = CommonState::new(Side::Client);
         common_state.set_max_fragment_size(config.max_fragment_size)?;
         common_state.protocol = proto;
-        #[cfg(feature = "secret_extraction")]
-        {
-            common_state.enable_secret_extraction = config.enable_secret_extraction;
-        }
+        common_state.enable_secret_extraction = config.enable_secret_extraction;
         let mut data = ClientConnectionData::new();
 
         let mut cx = hs::ClientContext {
@@ -671,6 +661,7 @@ impl ConnectionCore<ClientConnectionData> {
 }
 
 /// State associated with a client connection.
+#[derive(Debug)]
 pub struct ClientConnectionData {
     pub(super) early_data: EarlyData,
     pub(super) resumption_ciphersuite: Option<SupportedCipherSuite>,
