@@ -6,14 +6,14 @@ use crate::log::trace;
 use crate::msgs::deframer::{Deframed, MessageDeframer};
 use crate::msgs::handshake::Random;
 use crate::msgs::message::{Message, MessagePayload, PlainMessage};
-#[cfg(feature = "secret_extraction")]
 use crate::suites::{ExtractedSecrets, PartiallyExtractedSecrets};
 use crate::vecbuf::ChunkVecBuffer;
 
-use std::fmt::Debug;
+use alloc::boxed::Box;
+use core::fmt::Debug;
+use core::mem;
+use core::ops::{Deref, DerefMut};
 use std::io;
-use std::mem;
-use std::ops::{Deref, DerefMut};
 
 /// A client or server connection.
 #[derive(Debug)]
@@ -83,16 +83,6 @@ impl Connection {
         }
     }
 
-    /// Extract secrets, to set up kTLS for example
-    #[cfg(feature = "secret_extraction")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "secret_extraction")))]
-    pub fn extract_secrets(self) -> Result<ExtractedSecrets, Error> {
-        match self {
-            Self::Client(conn) => conn.extract_secrets(),
-            Self::Server(conn) => conn.extract_secrets(),
-        }
-    }
-
     /// This function uses `io` to complete any outstanding IO for this connection.
     ///
     /// See [`ConnectionCommon::complete_io()`] for more information.
@@ -106,6 +96,15 @@ impl Connection {
             Self::Server(conn) => conn.complete_io(io),
         }
     }
+
+    /// Extract secrets, so they can be used when configuring kTLS, for example.
+    /// Should be used with care as it exposes secret key material.
+    pub fn dangerous_extract_secrets(self) -> Result<ExtractedSecrets, Error> {
+        match self {
+            Self::Client(client) => client.dangerous_extract_secrets(),
+            Self::Server(server) => server.dangerous_extract_secrets(),
+        }
+    }
 }
 
 impl Deref for Connection {
@@ -143,8 +142,8 @@ impl<'a> io::Read for Reader<'a> {
     /// connection, so the underlying TCP connection should be half-closed too.
     ///
     /// If the peer closes the TLS session uncleanly (a TCP EOF without sending a
-    /// `close_notify` alert) this function returns `Err(ErrorKind::UnexpectedEof.into())`
-    /// once any pending data has been read.
+    /// `close_notify` alert) this function returns a `std::io::Error` of type
+    /// `ErrorKind::UnexpectedEof` once any pending data has been read.
     ///
     /// Note that support for `close_notify` varies in peer TLS libraries: many do not
     /// support it and uncleanly close the TCP connection (this might be
@@ -165,8 +164,13 @@ impl<'a> io::Read for Reader<'a> {
                 // cleanly closed; don't care about TCP EOF: express this as Ok(0)
                 (true, _) => {}
                 // unclean closure
-                (false, true) => return Err(io::ErrorKind::UnexpectedEof.into()),
-                // connection still going, but need more data: signal `WouldBlock` so that
+                (false, true) => {
+                    return Err(io::Error::new(
+                        io::ErrorKind::UnexpectedEof,
+                        UNEXPECTED_EOF_MESSAGE,
+                    ))
+                }
+                // connection still going, but needs more data: signal `WouldBlock` so that
                 // the caller knows this
                 (false, false) => return Err(io::ErrorKind::WouldBlock.into()),
             }
@@ -183,8 +187,8 @@ impl<'a> io::Read for Reader<'a> {
     /// should be half-closed too.
     ///
     /// If the peer closes the TLS session uncleanly (a TCP EOF without sending a
-    /// `close_notify` alert) this function returns `Err(ErrorKind::UnexpectedEof.into())`
-    /// once any pending data has been read.
+    /// `close_notify` alert) this function returns a `std::io::Error` of type
+    /// `ErrorKind::UnexpectedEof` once any pending data has been read.
     ///
     /// Note that support for `close_notify` varies in peer TLS libraries: many do not
     /// support it and uncleanly close the TCP connection (this might be
@@ -209,7 +213,12 @@ impl<'a> io::Read for Reader<'a> {
                 // cleanly closed; don't care about TCP EOF: express this as Ok(0)
                 (true, _) => {}
                 // unclean closure
-                (false, true) => return Err(io::ErrorKind::UnexpectedEof.into()),
+                (false, true) => {
+                    return Err(io::Error::new(
+                        io::ErrorKind::UnexpectedEof,
+                        UNEXPECTED_EOF_MESSAGE,
+                    ));
+                }
                 // connection still going, but need more data: signal `WouldBlock` so that
                 // the caller knows this
                 (false, false) => return Err(io::ErrorKind::WouldBlock.into()),
@@ -222,6 +231,9 @@ impl<'a> io::Read for Reader<'a> {
 
 /// Internal trait implemented by the [`ServerConnection`]/[`ClientConnection`]
 /// allowing them to be the subject of a [`Writer`].
+///
+/// [`ServerConnection`]: crate::ServerConnection
+/// [`ClientConnection`]: crate::ClientConnection
 pub(crate) trait PlaintextSink {
     fn write(&mut self, buf: &[u8]) -> io::Result<usize>;
     fn write_vectored(&mut self, bufs: &[io::IoSlice<'_>]) -> io::Result<usize>;
@@ -308,7 +320,7 @@ impl ConnectionRandoms {
 
 fn is_valid_ccs(msg: &PlainMessage) -> bool {
     // We passthrough ChangeCipherSpec messages in the deframer without decrypting them.
-    // nb. this is prior to the record layer, so is unencrypted. see
+    // Note: this is prior to the record layer, so is unencrypted. See
     // third paragraph of section 5 in RFC8446.
     msg.typ == ContentType::ChangeCipherSpec && msg.payload.0 == [0x01]
 }
@@ -537,6 +549,8 @@ impl<Data> ConnectionCommon<Data> {
     ///
     /// This function fails if called prior to the handshake completing;
     /// check with [`CommonState::is_handshaking`] first.
+    ///
+    /// This function fails if `output.len()` is zero.
     #[inline]
     pub fn export_keying_material<T: AsMut<[u8]>>(
         &self,
@@ -549,9 +563,8 @@ impl<Data> ConnectionCommon<Data> {
     }
 
     /// Extract secrets, so they can be used when configuring kTLS, for example.
-    #[cfg(feature = "secret_extraction")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "secret_extraction")))]
-    pub fn extract_secrets(self) -> Result<ExtractedSecrets, Error> {
+    /// Should be used with care as it exposes secret key material.
+    pub fn dangerous_extract_secrets(self) -> Result<ExtractedSecrets, Error> {
         if !self.enable_secret_extraction {
             return Err(Error::General("Secret extraction is disabled".into()));
         }
@@ -638,10 +651,10 @@ impl<Data> ConnectionCore<Data> {
 
     /// Pull a message out of the deframer and send any messages that need to be sent as a result.
     fn deframe(&mut self) -> Result<Option<PlainMessage>, Error> {
-        match self
-            .message_deframer
-            .pop(&mut self.common_state.record_layer)
-        {
+        match self.message_deframer.pop(
+            &mut self.common_state.record_layer,
+            self.common_state.negotiated_version,
+        ) {
             Ok(Some(Deframed {
                 want_close_before_decrypt,
                 aligned,
@@ -663,7 +676,6 @@ impl<Data> ConnectionCore<Data> {
             }
             Ok(None) => Ok(None),
             Err(err @ Error::InvalidMessage(_)) => {
-                #[cfg(feature = "quic")]
                 if self.common_state.is_quic() {
                     self.common_state.quic.alert = Some(AlertDescription::DecodeError);
                 }
@@ -740,6 +752,12 @@ impl<Data> ConnectionCore<Data> {
         label: &[u8],
         context: Option<&[u8]>,
     ) -> Result<T, Error> {
+        if output.as_mut().is_empty() {
+            return Err(Error::General(
+                "export_keying_material with zero-length output".into(),
+            ));
+        }
+
         match self.state.as_ref() {
             Ok(st) => st
                 .export_keying_material(output.as_mut(), label, context)
@@ -750,4 +768,7 @@ impl<Data> ConnectionCore<Data> {
 }
 
 /// Data specific to the peer's side (client or server).
-pub trait SideData {}
+pub trait SideData: Debug {}
+
+const UNEXPECTED_EOF_MESSAGE: &str = "peer closed connection without sending TLS close_notify: \
+https://docs.rs/rustls/latest/rustls/manual/_03_howto/index.html#unexpected-eof";
