@@ -1,5 +1,6 @@
+use alloc::vec::Vec;
+use core::ops::Range;
 use std::io;
-use std::ops::Range;
 
 use super::base::Payload;
 use super::codec::Codec;
@@ -39,7 +40,11 @@ impl MessageDeframer {
     /// Returns an `Error` if the deframer failed to parse some message contents or if decryption
     /// failed, `Ok(None)` if no full message is buffered or if trial decryption failed, and
     /// `Ok(Some(_))` if a valid message was found and decrypted successfully.
-    pub fn pop(&mut self, record_layer: &mut RecordLayer) -> Result<Option<Deframed>, Error> {
+    pub fn pop(
+        &mut self,
+        record_layer: &mut RecordLayer,
+        negotiated_version: Option<ProtocolVersion>,
+    ) -> Result<Option<Deframed>, Error> {
         if let Some(last_err) = self.last_error.clone() {
             return Err(last_err);
         } else if self.used == 0 {
@@ -87,10 +92,29 @@ impl MessageDeframer {
                 }
             };
 
-            // If we're in the middle of joining a handshake payload and the next message is not of
-            // type handshake, yield an error. Return CCS messages immediately without decrypting.
+            // Return CCS messages and early plaintext alerts immediately without decrypting.
             let end = start + rd.used();
-            if m.typ == ContentType::ChangeCipherSpec && self.joining_hs.is_none() {
+            let version_is_tls13 = matches!(negotiated_version, Some(ProtocolVersion::TLSv1_3));
+            let allowed_plaintext = match m.typ {
+                // CCS messages are always plaintext.
+                ContentType::ChangeCipherSpec => true,
+                // Alerts are allowed to be plaintext if-and-only-if:
+                // * The negotiated protocol version is TLS 1.3. - In TLS 1.2 it is unambiguous when
+                //   keying changes based on the CCS message. Only TLS 1.3 requires these heuristics.
+                // * We have not yet decrypted any messages from the peer - if we have we don't
+                //   expect any plaintext.
+                // * The payload size is indicative of a plaintext alert message.
+                ContentType::Alert
+                    if version_is_tls13
+                        && !record_layer.has_decrypted()
+                        && m.payload().len() <= 2 =>
+                {
+                    true
+                }
+                // In other circumstances, we expect all messages to be encrypted.
+                _ => false,
+            };
+            if self.joining_hs.is_none() && allowed_plaintext {
                 // This is unencrypted. We check the contents later.
                 self.discard(end);
                 return Ok(Some(Deframed {
@@ -135,7 +159,8 @@ impl MessageDeframer {
 
             // If it's not a handshake message, just return it -- no joining necessary.
             if msg.typ != ContentType::Handshake {
-                self.discard(start + rd.used());
+                let end = start + rd.used();
+                self.discard(end);
                 return Ok(Some(Deframed {
                     want_close_before_decrypt: false,
                     aligned: true,
@@ -195,8 +220,7 @@ impl MessageDeframer {
     }
 
     /// Allow pushing handshake messages directly into the buffer.
-    #[cfg(feature = "quic")]
-    pub fn push(&mut self, version: ProtocolVersion, payload: &[u8]) -> Result<(), Error> {
+    pub(crate) fn push(&mut self, version: ProtocolVersion, payload: &[u8]) -> Result<(), Error> {
         if self.used > 0 && self.joining_hs.is_none() {
             return Err(Error::General(
                 "cannot push QUIC messages into unrelated connection".into(),
@@ -407,17 +431,12 @@ fn payload_size(buf: &[u8]) -> Result<Option<usize>, Error> {
 
 #[derive(Debug)]
 pub struct Deframed {
-    pub want_close_before_decrypt: bool,
-    pub aligned: bool,
-    pub trial_decryption_finished: bool,
+    pub(crate) want_close_before_decrypt: bool,
+    pub(crate) aligned: bool,
+    pub(crate) trial_decryption_finished: bool,
     pub message: PlainMessage,
 }
 
-#[derive(Debug)]
-pub enum DeframerError {
-    HandshakePayloadSizeTooLarge,
-}
-
 const HEADER_SIZE: usize = 1 + 3;
 
 /// TLS allows for handshake messages of up to 16MB.  We
@@ -514,13 +533,21 @@ mod tests {
     }
 
     fn pop_first(d: &mut MessageDeframer, rl: &mut RecordLayer) {
-        let m = d.pop(rl).unwrap().unwrap().message;
+        let m = d
+            .pop(rl, None)
+            .unwrap()
+            .unwrap()
+            .message;
         assert_eq!(m.typ, ContentType::Handshake);
         Message::try_from(m).unwrap();
     }
 
     fn pop_second(d: &mut MessageDeframer, rl: &mut RecordLayer) {
-        let m = d.pop(rl).unwrap().unwrap().message;
+        let m = d
+            .pop(rl, None)
+            .unwrap()
+            .unwrap()
+            .message;
         assert_eq!(m.typ, ContentType::Alert);
         Message::try_from(m).unwrap();
     }
@@ -640,7 +667,7 @@ mod tests {
 
         let mut rl = RecordLayer::new();
         assert_eq!(
-            d.pop(&mut rl).unwrap_err(),
+            d.pop(&mut rl, None).unwrap_err(),
             Error::InvalidMessage(InvalidMessage::InvalidContentType)
         );
     }
@@ -655,7 +682,7 @@ mod tests {
 
         let mut rl = RecordLayer::new();
         assert_eq!(
-            d.pop(&mut rl).unwrap_err(),
+            d.pop(&mut rl, None).unwrap_err(),
             Error::InvalidMessage(InvalidMessage::UnknownProtocolVersion)
         );
     }
@@ -670,7 +697,7 @@ mod tests {
 
         let mut rl = RecordLayer::new();
         assert_eq!(
-            d.pop(&mut rl).unwrap_err(),
+            d.pop(&mut rl, None).unwrap_err(),
             Error::InvalidMessage(InvalidMessage::MessageTooLarge)
         );
     }
@@ -684,7 +711,11 @@ mod tests {
         );
 
         let mut rl = RecordLayer::new();
-        let m = d.pop(&mut rl).unwrap().unwrap().message;
+        let m = d
+            .pop(&mut rl, None)
+            .unwrap()
+            .unwrap()
+            .message;
         assert_eq!(m.typ, ContentType::ApplicationData);
         assert_eq!(m.payload.0.len(), 0);
         assert!(!d.has_pending());
@@ -701,12 +732,12 @@ mod tests {
 
         let mut rl = RecordLayer::new();
         assert_eq!(
-            d.pop(&mut rl).unwrap_err(),
+            d.pop(&mut rl, None).unwrap_err(),
             Error::InvalidMessage(InvalidMessage::InvalidEmptyPayload)
         );
         // CorruptMessage has been fused
         assert_eq!(
-            d.pop(&mut rl).unwrap_err(),
+            d.pop(&mut rl, None).unwrap_err(),
             Error::InvalidMessage(InvalidMessage::InvalidEmptyPayload)
         );
     }
