@@ -1,5 +1,4 @@
 use crate::enums::{CipherSuite, HandshakeType, ProtocolVersion, SignatureScheme};
-use crate::key::Certificate;
 use crate::msgs::base::{Payload, PayloadU16, PayloadU24, PayloadU8};
 use crate::msgs::codec::{put_u16, Codec, Reader};
 use crate::msgs::enums::{
@@ -7,19 +6,20 @@ use crate::msgs::enums::{
     KeyUpdateRequest, NamedGroup, PSKKeyExchangeMode, ServerNameType,
 };
 use crate::msgs::handshake::{
-    CertReqExtension, CertificateEntry, CertificateExtension, CertificatePayloadTLS13,
-    CertificateRequestPayload, CertificateRequestPayloadTLS13, CertificateStatus,
-    CertificateStatusRequest, ClientExtension, ClientHelloPayload, ClientSessionTicket,
-    ConvertProtocolNameList, ConvertServerNameList, DistinguishedName, ECDHEServerKeyExchange,
-    ECParameters, HandshakeMessagePayload, HandshakePayload, HasServerExtensions,
-    HelloRetryExtension, HelloRetryRequest, KeyShareEntry, NewSessionTicketExtension,
-    NewSessionTicketPayload, NewSessionTicketPayloadTLS13, PresharedKeyBinder,
-    PresharedKeyIdentity, PresharedKeyOffer, ProtocolName, Random, Sct, ServerECDHParams,
-    ServerExtension, ServerHelloPayload, ServerKeyExchangePayload, SessionId, UnknownExtension,
+    CertReqExtension, CertificateChain, CertificateEntry, CertificateExtension,
+    CertificatePayloadTls13, CertificateRequestPayload, CertificateRequestPayloadTls13,
+    CertificateStatus, CertificateStatusRequest, ClientExtension, ClientHelloPayload,
+    ClientSessionTicket, ConvertProtocolNameList, ConvertServerNameList, DistinguishedName,
+    EcParameters, EcdheServerKeyExchange, HandshakeMessagePayload, HandshakePayload,
+    HasServerExtensions, HelloRetryExtension, HelloRetryRequest, KeyShareEntry,
+    NewSessionTicketExtension, NewSessionTicketPayload, NewSessionTicketPayloadTls13,
+    PresharedKeyBinder, PresharedKeyIdentity, PresharedKeyOffer, ProtocolName, Random,
+    ServerEcdhParams, ServerExtension, ServerHelloPayload, ServerKeyExchangePayload, SessionId,
+    UnknownExtension,
 };
 use crate::verify::DigitallySignedStruct;
 
-use webpki::DnsNameRef;
+use pki_types::{CertificateDer, DnsName};
 
 #[test]
 fn rejects_short_random() {
@@ -74,8 +74,9 @@ fn accepts_short_sessionid() {
     let sess = SessionId::read(&mut rd).unwrap();
     println!("{:?}", sess);
 
+    #[cfg(feature = "tls12")]
     assert!(!sess.is_empty());
-    assert_eq!(sess.len(), 1);
+    assert_ne!(sess, SessionId::empty());
     assert!(!rd.any_left());
 }
 
@@ -86,8 +87,9 @@ fn accepts_empty_sessionid() {
     let sess = SessionId::read(&mut rd).unwrap();
     println!("{:?}", sess);
 
+    #[cfg(feature = "tls12")]
     assert!(sess.is_empty());
-    assert_eq!(sess.len(), 0);
+    assert_eq!(sess, SessionId::empty());
     assert!(!rd.any_left());
 }
 
@@ -132,7 +134,7 @@ fn refuses_server_ext_with_unparsed_bytes() {
 
 #[test]
 fn refuses_certificate_ext_with_unparsed_bytes() {
-    let bytes = [0x00u8, 0x12, 0x00, 0x03, 0x00, 0x00, 0x01];
+    let bytes = [0x00u8, 0x05, 0x00, 0x03, 0x00, 0x00, 0x01];
     let mut rd = Reader::init(&bytes);
     assert!(CertificateExtension::read(&mut rd).is_err());
 }
@@ -222,11 +224,8 @@ fn can_roundtrip_multiname_sni() {
 
             assert!(req.has_duplicate_names_for_type());
 
-            let dns_name_str: &str = req
-                .get_single_hostname()
-                .unwrap()
-                .into();
-            assert_eq!(dns_name_str, "hi");
+            let dns_name = req.get_single_hostname().unwrap();
+            assert_eq!(dns_name.as_ref(), "hi");
 
             assert_eq!(req[0].typ, ServerNameType::HostName);
             assert_eq!(req[1].typ, ServerNameType::HostName);
@@ -366,10 +365,10 @@ fn get_sample_clienthellopayload() -> ClientHelloPayload {
         cipher_suites: vec![CipherSuite::TLS_NULL_WITH_NULL_NULL],
         compression_methods: vec![Compression::Null],
         extensions: vec![
-            ClientExtension::ECPointFormats(ECPointFormat::SUPPORTED.to_vec()),
+            ClientExtension::EcPointFormats(ECPointFormat::SUPPORTED.to_vec()),
             ClientExtension::NamedGroups(vec![NamedGroup::X25519]),
             ClientExtension::SignatureAlgorithms(vec![SignatureScheme::ECDSA_NISTP256_SHA256]),
-            ClientExtension::make_sni(DnsNameRef::try_from_ascii_str("hello").unwrap()),
+            ClientExtension::make_sni(&DnsName::try_from("hello").unwrap()),
             ClientExtension::SessionTicket(ClientSessionTicket::Request),
             ClientExtension::SessionTicket(ClientSessionTicket::Offer(Payload(vec![]))),
             ClientExtension::Protocols(vec![ProtocolName::from(vec![0])]),
@@ -389,7 +388,6 @@ fn get_sample_clienthellopayload() -> ClientHelloPayload {
             ClientExtension::Cookie(PayloadU16(vec![1, 2, 3])),
             ClientExtension::ExtendedMasterSecretRequest,
             ClientExtension::CertificateStatusRequest(CertificateStatusRequest::build_ocsp()),
-            ClientExtension::SignedCertificateTimestampRequest,
             ClientExtension::TransportParameters(vec![1, 2, 3]),
             ClientExtension::Unknown(UnknownExtension {
                 typ: ExtensionType::Unknown(12345),
@@ -526,6 +524,7 @@ fn client_get_namedgroups_extension() {
     });
 }
 
+#[cfg(feature = "tls12")]
 #[test]
 fn client_get_ecpoints_extension() {
     test_client_extension_getter(ExtensionType::ECPointFormats, |chp| {
@@ -605,7 +604,7 @@ fn test_truncated_helloretry_extension_is_detected() {
 
 fn test_helloretry_extension_getter(typ: ExtensionType, getter: fn(&HelloRetryRequest) -> bool) {
     let mut hrr = get_sample_helloretryrequest();
-    let mut exts = std::mem::take(&mut hrr.extensions);
+    let mut exts = core::mem::take(&mut hrr.extensions);
     exts.retain(|ext| ext.get_type() == typ);
 
     assert!(!getter(&hrr));
@@ -708,11 +707,6 @@ fn server_get_ecpoints_extension() {
     });
 }
 
-#[test]
-fn server_get_sct_list() {
-    test_server_extension_getter(ExtensionType::SCT, |shp| shp.get_sct_list().is_some());
-}
-
 #[test]
 fn server_get_supported_versions() {
     test_server_extension_getter(ExtensionType::SupportedVersions, |shp| {
@@ -724,7 +718,7 @@ fn test_cert_extension_getter(typ: ExtensionType, getter: fn(&CertificateEntry)
     let mut ce = get_sample_certificatepayloadtls13()
         .entries
         .remove(0);
-    let mut exts = std::mem::take(&mut ce.exts);
+    let mut exts = core::mem::take(&mut ce.exts);
     exts.retain(|ext| ext.get_type() == typ);
 
     assert!(!getter(&ce));
@@ -746,11 +740,6 @@ fn certentry_get_ocsp_response() {
     });
 }
 
-#[test]
-fn certentry_get_scts() {
-    test_cert_extension_getter(ExtensionType::SCT, |ce| ce.get_scts().is_some());
-}
-
 fn get_sample_serverhellopayload() -> ServerHelloPayload {
     ServerHelloPayload {
         legacy_version: ProtocolVersion::TLSv1_2,
@@ -759,7 +748,7 @@ fn get_sample_serverhellopayload() -> ServerHelloPayload {
         cipher_suite: CipherSuite::TLS_NULL_WITH_NULL_NULL,
         compression_method: Compression::Null,
         extensions: vec![
-            ServerExtension::ECPointFormats(ECPointFormat::SUPPORTED.to_vec()),
+            ServerExtension::EcPointFormats(ECPointFormat::SUPPORTED.to_vec()),
             ServerExtension::ServerNameAck,
             ServerExtension::SessionTicketAck,
             ServerExtension::RenegotiationInfo(PayloadU8(vec![0])),
@@ -768,7 +757,6 @@ fn get_sample_serverhellopayload() -> ServerHelloPayload {
             ServerExtension::PresharedKey(3),
             ServerExtension::ExtendedMasterSecretAck,
             ServerExtension::CertificateStatusAck,
-            ServerExtension::SignedCertificateTimestamp(vec![Sct::from(vec![0])]),
             ServerExtension::SupportedVersions(ProtocolVersion::TLSv1_2),
             ServerExtension::TransportParameters(vec![1, 2, 3]),
             ServerExtension::Unknown(UnknownExtension {
@@ -806,16 +794,15 @@ fn get_sample_helloretryrequest() -> HelloRetryRequest {
     }
 }
 
-fn get_sample_certificatepayloadtls13() -> CertificatePayloadTLS13 {
-    CertificatePayloadTLS13 {
+fn get_sample_certificatepayloadtls13() -> CertificatePayloadTls13 {
+    CertificatePayloadTls13 {
         context: PayloadU8(vec![1, 2, 3]),
         entries: vec![CertificateEntry {
-            cert: Certificate(vec![3, 4, 5]),
+            cert: CertificateDer::from(vec![3, 4, 5]),
             exts: vec![
                 CertificateExtension::CertificateStatus(CertificateStatus {
                     ocsp_response: PayloadU24(vec![1, 2, 3]),
                 }),
-                CertificateExtension::SignedCertificateTimestamp(vec![Sct::from(vec![0])]),
                 CertificateExtension::Unknown(UnknownExtension {
                     typ: ExtensionType::Unknown(12345),
                     payload: Payload(vec![1, 2, 3]),
@@ -826,9 +813,9 @@ fn get_sample_certificatepayloadtls13() -> CertificatePayloadTLS13 {
 }
 
 fn get_sample_serverkeyexchangepayload_ecdhe() -> ServerKeyExchangePayload {
-    ServerKeyExchangePayload::ECDHE(ECDHEServerKeyExchange {
-        params: ServerECDHParams {
-            curve_params: ECParameters {
+    ServerKeyExchangePayload::Ecdhe(EcdheServerKeyExchange {
+        params: ServerEcdhParams {
+            curve_params: EcParameters {
                 curve_type: ECCurveType::NamedCurve,
                 named_group: NamedGroup::X25519,
             },
@@ -850,8 +837,8 @@ fn get_sample_certificaterequestpayload() -> CertificateRequestPayload {
     }
 }
 
-fn get_sample_certificaterequestpayloadtls13() -> CertificateRequestPayloadTLS13 {
-    CertificateRequestPayloadTLS13 {
+fn get_sample_certificaterequestpayloadtls13() -> CertificateRequestPayloadTls13 {
+    CertificateRequestPayloadTls13 {
         context: PayloadU8(vec![1, 2, 3]),
         extensions: vec![
             CertReqExtension::SignatureAlgorithms(vec![SignatureScheme::ECDSA_NISTP256_SHA256]),
@@ -871,8 +858,8 @@ fn get_sample_newsessionticketpayload() -> NewSessionTicketPayload {
     }
 }
 
-fn get_sample_newsessionticketpayloadtls13() -> NewSessionTicketPayloadTLS13 {
-    NewSessionTicketPayloadTLS13 {
+fn get_sample_newsessionticketpayloadtls13() -> NewSessionTicketPayloadTls13 {
+    NewSessionTicketPayloadTls13 {
         lifetime: 123,
         age_add: 1234,
         nonce: PayloadU8(vec![1, 2, 3]),
@@ -914,7 +901,9 @@ fn get_all_tls12_handshake_payloads() -> Vec<HandshakeMessagePayload> {
         },
         HandshakeMessagePayload {
             typ: HandshakeType::Certificate,
-            payload: HandshakePayload::Certificate(vec![Certificate(vec![1, 2, 3])]),
+            payload: HandshakePayload::Certificate(CertificateChain(vec![CertificateDer::from(
+                vec![1, 2, 3],
+            )])),
         },
         HandshakeMessagePayload {
             typ: HandshakeType::ServerKeyExchange,
@@ -1042,7 +1031,7 @@ fn get_all_tls13_handshake_payloads() -> Vec<HandshakeMessagePayload> {
         },
         HandshakeMessagePayload {
             typ: HandshakeType::Certificate,
-            payload: HandshakePayload::CertificateTLS13(get_sample_certificatepayloadtls13()),
+            payload: HandshakePayload::CertificateTls13(get_sample_certificatepayloadtls13()),
         },
         HandshakeMessagePayload {
             typ: HandshakeType::ServerKeyExchange,
@@ -1058,7 +1047,7 @@ fn get_all_tls13_handshake_payloads() -> Vec<HandshakeMessagePayload> {
         },
         HandshakeMessagePayload {
             typ: HandshakeType::CertificateRequest,
-            payload: HandshakePayload::CertificateRequestTLS13(
+            payload: HandshakePayload::CertificateRequestTls13(
                 get_sample_certificaterequestpayloadtls13(),
             ),
         },
@@ -1079,7 +1068,7 @@ fn get_all_tls13_handshake_payloads() -> Vec<HandshakeMessagePayload> {
         },
         HandshakeMessagePayload {
             typ: HandshakeType::NewSessionTicket,
-            payload: HandshakePayload::NewSessionTicketTLS13(
+            payload: HandshakePayload::NewSessionTicketTls13(
                 get_sample_newsessionticketpayloadtls13(),
             ),
         },
@@ -1212,3 +1201,12 @@ fn can_decode_server_hello_from_api_devicecheck_apple_com() {
     let hm = HandshakeMessagePayload::read(&mut r).unwrap();
     println!("msg: {:?}", hm);
 }
+
+#[test]
+fn wrapped_dn_encoding() {
+    let subject = b"subject";
+    let dn = DistinguishedName::in_sequence(&subject[..]);
+    const DER_SEQUENCE_TAG: u8 = 0x30;
+    let expected_prefix = vec![DER_SEQUENCE_TAG, subject.len() as u8];
+    assert_eq!(dn.as_ref(), [expected_prefix, subject.to_vec()].concat());
+}
