@@ -1,42 +1,52 @@
 use crate::check::inappropriate_message;
 use crate::common_state::{CommonState, Side, State};
 use crate::conn::ConnectionRandoms;
+use crate::crypto::ActiveKeyExchange;
 use crate::enums::ProtocolVersion;
 use crate::enums::{AlertDescription, ContentType, HandshakeType};
 use crate::error::{Error, PeerIncompatible, PeerMisbehaved};
 use crate::hash_hs::HandshakeHash;
-use crate::key::Certificate;
 #[cfg(feature = "logging")]
 use crate::log::{debug, trace};
 use crate::msgs::base::Payload;
 use crate::msgs::ccs::ChangeCipherSpecPayload;
 use crate::msgs::codec::Codec;
-use crate::msgs::handshake::{ClientECDHParams, HandshakeMessagePayload, HandshakePayload};
+use crate::msgs::handshake::{
+    CertificateChain, ClientEcdhParams, HandshakeMessagePayload, HandshakePayload,
+};
 use crate::msgs::handshake::{NewSessionTicketPayload, SessionId};
 use crate::msgs::message::{Message, MessagePayload};
 use crate::msgs::persist;
-#[cfg(feature = "secret_extraction")]
 use crate::suites::PartiallyExtractedSecrets;
 use crate::tls12::{self, ConnectionSecrets, Tls12CipherSuite};
-use crate::{kx, ticketer, verify};
+use crate::verify;
 
 use super::common::ActiveCertifiedKey;
 use super::hs::{self, ServerContext};
 use super::server_conn::{ProducesTickets, ServerConfig, ServerConnectionData};
 
-use ring::constant_time;
+use pki_types::UnixTime;
+use subtle::ConstantTimeEq;
 
-use std::sync::Arc;
+use alloc::borrow::ToOwned;
+use alloc::boxed::Box;
+use alloc::string::ToString;
+use alloc::sync::Arc;
+use alloc::vec;
+use alloc::vec::Vec;
 
 pub(super) use client_hello::CompleteClientHelloHandling;
 
 mod client_hello {
+    use pki_types::CertificateDer;
+
+    use crate::crypto::SupportedKxGroup;
     use crate::enums::SignatureScheme;
     use crate::msgs::enums::ECPointFormat;
     use crate::msgs::enums::{ClientCertificateType, Compression};
-    use crate::msgs::handshake::ServerECDHParams;
+    use crate::msgs::handshake::{CertificateChain, ServerEcdhParams};
     use crate::msgs::handshake::{CertificateRequestPayload, ClientSessionTicket, Random};
-    use crate::msgs::handshake::{CertificateStatus, ECDHEServerKeyExchange};
+    use crate::msgs::handshake::{CertificateStatus, EcdheServerKeyExchange};
     use crate::msgs::handshake::{ClientExtension, SessionId};
     use crate::msgs::handshake::{ClientHelloPayload, ServerHelloPayload};
     use crate::msgs::handshake::{ServerExtension, ServerKeyExchangePayload};
@@ -76,21 +86,28 @@ mod client_hello {
             let groups_ext = client_hello
                 .get_namedgroups_extension()
                 .ok_or_else(|| {
-                    hs::incompatible(cx.common, PeerIncompatible::NamedGroupsExtensionRequired)
+                    cx.common.send_fatal_alert(
+                        AlertDescription::HandshakeFailure,
+                        PeerIncompatible::NamedGroupsExtensionRequired,
+                    )
                 })?;
             let ecpoints_ext = client_hello
                 .get_ecpoints_extension()
                 .ok_or_else(|| {
-                    hs::incompatible(cx.common, PeerIncompatible::EcPointsExtensionRequired)
+                    cx.common.send_fatal_alert(
+                        AlertDescription::HandshakeFailure,
+                        PeerIncompatible::EcPointsExtensionRequired,
+                    )
                 })?;
 
             trace!("namedgroups {:?}", groups_ext);
             trace!("ecpoints {:?}", ecpoints_ext);
 
             if !ecpoints_ext.contains(&ECPointFormat::Uncompressed) {
-                cx.common
-                    .send_fatal_alert(AlertDescription::IllegalParameter);
-                return Err(PeerIncompatible::UncompressedEcPointsRequired.into());
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::IllegalParameter,
+                    PeerIncompatible::UncompressedEcPointsRequired,
+                ));
             }
 
             // -- If TLS1.3 is enabled, signal the downgrade in the server random
@@ -154,38 +171,46 @@ mod client_hello {
                 .resolve_sig_schemes(&sigschemes_ext);
 
             if sigschemes.is_empty() {
-                return Err(hs::incompatible(
-                    cx.common,
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::HandshakeFailure,
                     PeerIncompatible::NoSignatureSchemesInCommon,
                 ));
             }
 
             let group = self
                 .config
+                .provider
                 .kx_groups
                 .iter()
-                .find(|skxg| groups_ext.contains(&skxg.name))
+                .find(|skxg| groups_ext.contains(&skxg.name()))
                 .cloned()
-                .ok_or_else(|| hs::incompatible(cx.common, PeerIncompatible::NoKxGroupsInCommon))?;
+                .ok_or_else(|| {
+                    cx.common.send_fatal_alert(
+                        AlertDescription::HandshakeFailure,
+                        PeerIncompatible::NoKxGroupsInCommon,
+                    )
+                })?;
 
             let ecpoint = ECPointFormat::SUPPORTED
                 .iter()
                 .find(|format| ecpoints_ext.contains(format))
                 .cloned()
                 .ok_or_else(|| {
-                    hs::incompatible(cx.common, PeerIncompatible::NoEcPointFormatsInCommon)
+                    cx.common.send_fatal_alert(
+                        AlertDescription::HandshakeFailure,
+                        PeerIncompatible::NoEcPointFormatsInCommon,
+                    )
                 })?;
 
             debug_assert_eq!(ecpoint, ECPointFormat::Uncompressed);
 
-            let (mut ocsp_response, mut sct_list) =
-                (server_key.get_ocsp(), server_key.get_sct_list());
+            let mut ocsp_response = server_key.get_ocsp();
 
             // If we're not offered a ticket or a potential session ID, allocate a session ID.
             if !self.config.session_storage.can_cache() {
                 self.session_id = SessionId::empty();
             } else if self.session_id.is_empty() && !ticket_received {
-                self.session_id = SessionId::random()?;
+                self.session_id = SessionId::random(self.config.provider.secure_random)?;
             }
 
             self.send_ticket = emit_server_hello(
@@ -196,7 +221,6 @@ mod client_hello {
                 self.suite,
                 self.using_ems,
                 &mut ocsp_response,
-                &mut sct_list,
                 client_hello,
                 None,
                 &self.randoms,
@@ -253,9 +277,10 @@ mod client_hello {
             debug!("Resuming connection");
 
             if resumedata.extended_ms && !self.using_ems {
-                return Err(cx
-                    .common
-                    .illegal_param(PeerMisbehaved::ResumptionAttemptedWithVariedEms));
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::IllegalParameter,
+                    PeerMisbehaved::ResumptionAttemptedWithVariedEms,
+                ));
             }
 
             self.session_id = *id;
@@ -267,7 +292,6 @@ mod client_hello {
                 self.suite,
                 self.using_ems,
                 &mut None,
-                &mut None,
                 client_hello,
                 Some(&resumedata),
                 &self.randoms,
@@ -323,22 +347,13 @@ mod client_hello {
         suite: &'static Tls12CipherSuite,
         using_ems: bool,
         ocsp_response: &mut Option<&[u8]>,
-        sct_list: &mut Option<&[u8]>,
         hello: &ClientHelloPayload,
         resumedata: Option<&persist::ServerSessionValue>,
         randoms: &ConnectionRandoms,
         extra_exts: Vec<ServerExtension>,
     ) -> Result<bool, Error> {
         let mut ep = hs::ExtensionProcessing::new();
-        ep.process_common(
-            config,
-            cx,
-            ocsp_response,
-            sct_list,
-            hello,
-            resumedata,
-            extra_exts,
-        )?;
+        ep.process_common(config, cx, ocsp_response, hello, resumedata, extra_exts)?;
         ep.process_tls12(config, hello, using_ems);
 
         let sh = Message {
@@ -365,13 +380,13 @@ mod client_hello {
     fn emit_certificate(
         transcript: &mut HandshakeHash,
         common: &mut CommonState,
-        cert_chain: &[Certificate],
+        cert_chain: &[CertificateDer<'static>],
     ) {
         let c = Message {
             version: ProtocolVersion::TLSv1_2,
             payload: MessagePayload::handshake(HandshakeMessagePayload {
                 typ: HandshakeType::Certificate,
-                payload: HandshakePayload::Certificate(cert_chain.to_owned()),
+                payload: HandshakePayload::Certificate(CertificateChain(cert_chain.to_vec())),
             }),
         };
 
@@ -398,12 +413,14 @@ mod client_hello {
         transcript: &mut HandshakeHash,
         common: &mut CommonState,
         sigschemes: Vec<SignatureScheme>,
-        skxg: &'static kx::SupportedKxGroup,
+        selected_group: &'static dyn SupportedKxGroup,
         signing_key: &dyn sign::SigningKey,
         randoms: &ConnectionRandoms,
-    ) -> Result<kx::KeyExchange, Error> {
-        let kx = kx::KeyExchange::start(skxg).ok_or(Error::FailedToGetRandomBytes)?;
-        let secdh = ServerECDHParams::new(skxg.name, kx.pubkey.as_ref());
+    ) -> Result<Box<dyn ActiveKeyExchange>, Error> {
+        let kx = selected_group
+            .start()
+            .map_err(|_| Error::FailedToGetRandomBytes)?;
+        let secdh = ServerEcdhParams::new(&*kx);
 
         let mut msg = Vec::new();
         msg.extend(randoms.client);
@@ -416,7 +433,7 @@ mod client_hello {
         let sigscheme = signer.scheme();
         let sig = signer.sign(&msg)?;
 
-        let skx = ServerKeyExchangePayload::ECDHE(ECDHEServerKeyExchange {
+        let skx = ServerKeyExchangePayload::Ecdhe(EcdheServerKeyExchange {
             params: secdh,
             dss: DigitallySignedStruct::new(sigscheme, sig),
         });
@@ -449,7 +466,7 @@ mod client_hello {
 
         let names = config
             .verifier
-            .client_auth_root_subjects()
+            .root_hint_subjects()
             .to_vec();
 
         let cr = CertificateRequestPayload {
@@ -497,7 +514,7 @@ struct ExpectCertificate {
     session_id: SessionId,
     suite: &'static Tls12CipherSuite,
     using_ems: bool,
-    server_kx: kx::KeyExchange,
+    server_kx: Box<dyn ActiveKeyExchange>,
     send_ticket: bool,
 }
 
@@ -520,9 +537,10 @@ impl State<ServerConnectionData> for ExpectCertificate {
 
         let client_cert = match cert_chain.split_first() {
             None if mandatory => {
-                cx.common
-                    .send_fatal_alert(AlertDescription::CertificateRequired);
-                return Err(Error::NoCertificatesPresented);
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::CertificateRequired,
+                    Error::NoCertificatesPresented,
+                ));
             }
             None => {
                 debug!("client auth requested but no certificate supplied");
@@ -530,10 +548,9 @@ impl State<ServerConnectionData> for ExpectCertificate {
                 None
             }
             Some((end_entity, intermediates)) => {
-                let now = std::time::SystemTime::now();
                 self.config
                     .verifier
-                    .verify_client_cert(end_entity, intermediates, now)
+                    .verify_client_cert(end_entity, intermediates, UnixTime::now())
                     .map_err(|err| {
                         cx.common
                             .send_cert_verify_error_alert(err)
@@ -565,8 +582,8 @@ struct ExpectClientKx {
     session_id: SessionId,
     suite: &'static Tls12CipherSuite,
     using_ems: bool,
-    server_kx: kx::KeyExchange,
-    client_cert: Option<Vec<Certificate>>,
+    server_kx: Box<dyn ActiveKeyExchange>,
+    client_cert: Option<CertificateChain>,
     send_ticket: bool,
 }
 
@@ -585,7 +602,7 @@ impl State<ServerConnectionData> for ExpectClientKx {
         // Complete key agreement, and set up encryption with the
         // resulting premaster secret.
         let peer_kx_params =
-            tls12::decode_ecdh_params::<ClientECDHParams>(cx.common, &client_kx.0)?;
+            tls12::decode_ecdh_params::<ClientEcdhParams>(cx.common, &client_kx.0)?;
         let secrets = ConnectionSecrets::from_key_exchange(
             self.server_kx,
             &peer_kx_params.public.0,
@@ -633,7 +650,7 @@ struct ExpectCertificateVerify {
     transcript: HandshakeHash,
     session_id: SessionId,
     using_ems: bool,
-    client_cert: Vec<Certificate>,
+    client_cert: CertificateChain,
     send_ticket: bool,
 }
 
@@ -659,9 +676,10 @@ impl State<ServerConnectionData> for ExpectCertificateVerify {
                     // `transcript.abandon_client_auth()` can extract it, but its only caller in
                     // this flow will also set `ExpectClientKx::client_cert` to `None`, making it
                     // impossible to reach this state.
-                    cx.common
-                        .send_fatal_alert(AlertDescription::AccessDenied);
-                    return Err(Error::General("client authentication not set up".into()));
+                    return Err(cx.common.send_fatal_alert(
+                        AlertDescription::AccessDenied,
+                        Error::General("client authentication not set up".into()),
+                    ));
                 }
             }
         };
@@ -735,16 +753,15 @@ fn get_server_connection_value_tls12(
     secrets: &ConnectionSecrets,
     using_ems: bool,
     cx: &ServerContext<'_>,
-    time_now: ticketer::TimeBase,
+    time_now: UnixTime,
 ) -> persist::ServerSessionValue {
     let version = ProtocolVersion::TLSv1_2;
-    let secret = secrets.get_master_secret();
 
     let mut v = persist::ServerSessionValue::new(
         cx.data.sni.as_ref(),
         version,
         secrets.suite().common.suite,
-        secret,
+        secrets.master_secret(),
         cx.common.peer_certificates.clone(),
         cx.common.alpn_protocol.clone(),
         cx.data.resumption_data.clone(),
@@ -766,8 +783,8 @@ fn emit_ticket(
     cx: &mut ServerContext<'_>,
     ticketer: &dyn ProducesTickets,
 ) -> Result<(), Error> {
-    let time_now = ticketer::TimeBase::now()?;
-    let plain = get_server_connection_value_tls12(secrets, using_ems, cx, time_now).get_encoding();
+    let plain =
+        get_server_connection_value_tls12(secrets, using_ems, cx, UnixTime::now()).get_encoding();
 
     // If we can't produce a ticket for some reason, we can't
     // report an error. Send an empty one.
@@ -842,20 +859,24 @@ impl State<ServerConnectionData> for ExpectFinished {
         let vh = self.transcript.get_current_hash();
         let expect_verify_data = self.secrets.client_verify_data(&vh);
 
-        let _fin_verified =
-            constant_time::verify_slices_are_equal(&expect_verify_data, &finished.0)
-                .map_err(|_| {
-                    cx.common
-                        .send_fatal_alert(AlertDescription::DecryptError);
-                    Error::DecryptError
-                })
-                .map(|_| verify::FinishedMessageVerified::assertion())?;
+        let _fin_verified = match ConstantTimeEq::ct_eq(&expect_verify_data[..], &finished.0).into()
+        {
+            true => verify::FinishedMessageVerified::assertion(),
+            false => {
+                return Err(cx
+                    .common
+                    .send_fatal_alert(AlertDescription::DecryptError, Error::DecryptError));
+            }
+        };
 
         // Save connection, perhaps
         if !self.resuming && !self.session_id.is_empty() {
-            let time_now = ticketer::TimeBase::now()?;
-            let value =
-                get_server_connection_value_tls12(&self.secrets, self.using_ems, cx, time_now);
+            let value = get_server_connection_value_tls12(
+                &self.secrets,
+                self.using_ems,
+                cx,
+                UnixTime::now(),
+            );
 
             let worked = self
                 .config
@@ -930,7 +951,6 @@ impl State<ServerConnectionData> for ExpectTraffic {
         Ok(())
     }
 
-    #[cfg(feature = "secret_extraction")]
     fn extract_secrets(&self) -> Result<PartiallyExtractedSecrets, Error> {
         self.secrets
             .extract_secrets(Side::Server)
