@@ -1,31 +1,27 @@
 use crate::check::inappropriate_handshake_message;
-#[cfg(feature = "quic")]
 use crate::check::inappropriate_message;
-#[cfg(feature = "quic")]
 use crate::common_state::Protocol;
-#[cfg(feature = "secret_extraction")]
-use crate::common_state::Side;
-use crate::common_state::{CommonState, State};
+use crate::common_state::{CommonState, Side, State};
 use crate::conn::ConnectionRandoms;
 use crate::enums::ProtocolVersion;
 use crate::enums::{AlertDescription, ContentType, HandshakeType};
 use crate::error::{Error, PeerIncompatible, PeerMisbehaved};
 use crate::hash_hs::HandshakeHash;
-use crate::key::Certificate;
 #[cfg(feature = "logging")]
 use crate::log::{debug, trace, warn};
 use crate::msgs::codec::Codec;
 use crate::msgs::enums::KeyUpdateRequest;
+use crate::msgs::handshake::CertificateChain;
 use crate::msgs::handshake::HandshakeMessagePayload;
 use crate::msgs::handshake::HandshakePayload;
-use crate::msgs::handshake::{NewSessionTicketExtension, NewSessionTicketPayloadTLS13};
+use crate::msgs::handshake::{NewSessionTicketExtension, NewSessionTicketPayloadTls13};
 use crate::msgs::message::{Message, MessagePayload};
 use crate::msgs::persist;
 use crate::rand;
 use crate::server::ServerConfig;
-#[cfg(feature = "secret_extraction")]
 use crate::suites::PartiallyExtractedSecrets;
-use crate::ticketer;
+use crate::tls13::construct_client_verify_message;
+use crate::tls13::construct_server_verify_message;
 use crate::tls13::key_schedule::{KeyScheduleTraffic, KeyScheduleTrafficWithClientFinishedPending};
 use crate::tls13::Tls13CipherSuite;
 use crate::verify;
@@ -33,15 +29,20 @@ use crate::verify;
 use super::hs::{self, HandshakeHashOrBuffer, ServerContext};
 use super::server_conn::ServerConnectionData;
 
-use std::sync::Arc;
+use alloc::borrow::ToOwned;
+use alloc::boxed::Box;
+use alloc::sync::Arc;
+use alloc::vec;
+use alloc::vec::Vec;
 
-use ring::constant_time;
+use pki_types::{CertificateDer, UnixTime};
+use subtle::ConstantTimeEq;
 
 pub(super) use client_hello::CompleteClientHelloHandling;
 
 mod client_hello {
+    use crate::crypto::SupportedKxGroup;
     use crate::enums::SignatureScheme;
-    use crate::kx;
     use crate::msgs::base::{Payload, PayloadU8};
     use crate::msgs::ccs::ChangeCipherSpecPayload;
     use crate::msgs::enums::NamedGroup;
@@ -49,8 +50,8 @@ mod client_hello {
     use crate::msgs::handshake::CertReqExtension;
     use crate::msgs::handshake::CertificateEntry;
     use crate::msgs::handshake::CertificateExtension;
-    use crate::msgs::handshake::CertificatePayloadTLS13;
-    use crate::msgs::handshake::CertificateRequestPayloadTLS13;
+    use crate::msgs::handshake::CertificatePayloadTls13;
+    use crate::msgs::handshake::CertificateRequestPayloadTls13;
     use crate::msgs::handshake::CertificateStatus;
     use crate::msgs::handshake::ClientHelloPayload;
     use crate::msgs::handshake::HelloRetryExtension;
@@ -124,7 +125,7 @@ mod client_hello {
             let real_binder =
                 key_schedule.resumption_psk_binder_key_and_sign_verify_data(&handshake_hash);
 
-            constant_time::verify_slices_are_equal(real_binder.as_ref(), binder).is_ok()
+            ConstantTimeEq::ct_eq(real_binder.as_ref(), binder).into()
         }
 
         fn attempt_tls13_ticket_decryption(
@@ -153,78 +154,93 @@ mod client_hello {
             mut sigschemes_ext: Vec<SignatureScheme>,
         ) -> hs::NextStateOrError {
             if client_hello.compression_methods.len() != 1 {
-                return Err(cx
-                    .common
-                    .illegal_param(PeerMisbehaved::OfferedIncorrectCompressions));
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::IllegalParameter,
+                    PeerMisbehaved::OfferedIncorrectCompressions,
+                ));
             }
 
             let groups_ext = client_hello
                 .get_namedgroups_extension()
                 .ok_or_else(|| {
-                    hs::incompatible(cx.common, PeerIncompatible::NamedGroupsExtensionRequired)
+                    cx.common.send_fatal_alert(
+                        AlertDescription::HandshakeFailure,
+                        PeerIncompatible::NamedGroupsExtensionRequired,
+                    )
                 })?;
 
-            let tls13_schemes = sign::supported_sign_tls13();
-            sigschemes_ext.retain(|scheme| tls13_schemes.contains(scheme));
+            sigschemes_ext.retain(SignatureScheme::supported_in_tls13);
 
             let shares_ext = client_hello
                 .get_keyshare_extension()
                 .ok_or_else(|| {
-                    hs::incompatible(cx.common, PeerIncompatible::KeyShareExtensionRequired)
+                    cx.common.send_fatal_alert(
+                        AlertDescription::HandshakeFailure,
+                        PeerIncompatible::KeyShareExtensionRequired,
+                    )
                 })?;
 
             if client_hello.has_keyshare_extension_with_duplicates() {
-                return Err(cx
-                    .common
-                    .illegal_param(PeerMisbehaved::OfferedDuplicateKeyShares));
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::IllegalParameter,
+                    PeerMisbehaved::OfferedDuplicateKeyShares,
+                ));
             }
 
             let early_data_requested = client_hello.early_data_extension_offered();
 
             // EarlyData extension is illegal in second ClientHello
             if self.done_retry && early_data_requested {
-                return Err(cx
-                    .common
-                    .illegal_param(PeerMisbehaved::EarlyDataAttemptedInSecondClientHello));
+                return Err({
+                    cx.common.send_fatal_alert(
+                        AlertDescription::IllegalParameter,
+                        PeerMisbehaved::EarlyDataAttemptedInSecondClientHello,
+                    )
+                });
             }
 
             // choose a share that we support
-            let chosen_share = self
+            let chosen_share_and_kxg = self
                 .config
+                .provider
                 .kx_groups
                 .iter()
                 .find_map(|group| {
                     shares_ext
                         .iter()
-                        .find(|share| share.group == group.name)
+                        .find(|share| share.group == group.name())
+                        .map(|share| (share, *group))
                 });
 
-            let chosen_share = match chosen_share {
+            let chosen_share_and_kxg = match chosen_share_and_kxg {
                 Some(s) => s,
                 None => {
                     // We don't have a suitable key share.  Choose a suitable group and
                     // send a HelloRetryRequest.
                     let retry_group_maybe = self
                         .config
+                        .provider
                         .kx_groups
                         .iter()
-                        .find(|group| groups_ext.contains(&group.name))
+                        .find(|group| groups_ext.contains(&group.name()))
                         .cloned();
 
                     self.transcript.add_message(chm);
 
                     if let Some(group) = retry_group_maybe {
                         if self.done_retry {
-                            return Err(cx
-                                .common
-                                .illegal_param(PeerMisbehaved::RefusedToFollowHelloRetryRequest));
+                            return Err(cx.common.send_fatal_alert(
+                                AlertDescription::IllegalParameter,
+                                PeerMisbehaved::RefusedToFollowHelloRetryRequest,
+                            ));
                         }
 
                         emit_hello_retry_request(
                             &mut self.transcript,
                             self.suite,
+                            client_hello.session_id,
                             cx.common,
-                            group.name,
+                            group.name(),
                         );
                         emit_fake_ccs(cx.common);
 
@@ -252,8 +268,8 @@ mod client_hello {
                         };
                     }
 
-                    return Err(hs::incompatible(
-                        cx.common,
+                    return Err(cx.common.send_fatal_alert(
+                        AlertDescription::HandshakeFailure,
                         PeerIncompatible::NoKxGroupsInCommon,
                     ));
                 }
@@ -261,33 +277,45 @@ mod client_hello {
 
             let mut chosen_psk_index = None;
             let mut resumedata = None;
-            let time_now = ticketer::TimeBase::now()?;
 
             if let Some(psk_offer) = client_hello.get_psk() {
                 if !client_hello.check_psk_ext_is_last() {
-                    return Err(cx
-                        .common
-                        .illegal_param(PeerMisbehaved::PskExtensionMustBeLast));
+                    return Err(cx.common.send_fatal_alert(
+                        AlertDescription::IllegalParameter,
+                        PeerMisbehaved::PskExtensionMustBeLast,
+                    ));
+                }
+
+                // "A client MUST provide a "psk_key_exchange_modes" extension if it
+                //  offers a "pre_shared_key" extension. If clients offer
+                //  "pre_shared_key" without a "psk_key_exchange_modes" extension,
+                //  servers MUST abort the handshake." - RFC8446 4.2.9
+                if client_hello.get_psk_modes().is_none() {
+                    return Err(cx.common.send_fatal_alert(
+                        AlertDescription::MissingExtension,
+                        PeerMisbehaved::MissingPskModesExtension,
+                    ));
                 }
 
                 if psk_offer.binders.is_empty() {
-                    return Err(hs::decode_error(
-                        cx.common,
+                    return Err(cx.common.send_fatal_alert(
+                        AlertDescription::DecodeError,
                         PeerMisbehaved::MissingBinderInPskExtension,
                     ));
                 }
 
                 if psk_offer.binders.len() != psk_offer.identities.len() {
-                    return Err(cx
-                        .common
-                        .illegal_param(PeerMisbehaved::PskExtensionWithMismatchedIdsAndBinders));
+                    return Err(cx.common.send_fatal_alert(
+                        AlertDescription::IllegalParameter,
+                        PeerMisbehaved::PskExtensionWithMismatchedIdsAndBinders,
+                    ));
                 }
 
                 for (i, psk_id) in psk_offer.identities.iter().enumerate() {
                     let resume = match self
                         .attempt_tls13_ticket_decryption(&psk_id.identity.0)
                         .map(|resumedata| {
-                            resumedata.set_freshness(psk_id.obfuscated_ticket_age, time_now)
+                            resumedata.set_freshness(psk_id.obfuscated_ticket_age, UnixTime::now())
                         })
                         .filter(|resumedata| {
                             hs::can_resume(self.suite.into(), &cx.data.sni, false, resumedata)
@@ -302,9 +330,10 @@ mod client_hello {
                         &resume.master_secret.0,
                         psk_offer.binders[i].as_ref(),
                     ) {
-                        cx.common
-                            .send_fatal_alert(AlertDescription::DecryptError);
-                        return Err(PeerMisbehaved::IncorrectBinder.into());
+                        return Err(cx.common.send_fatal_alert(
+                            AlertDescription::DecryptError,
+                            PeerMisbehaved::IncorrectBinder,
+                        ));
                     }
 
                     chosen_psk_index = Some(i);
@@ -335,7 +364,7 @@ mod client_hello {
                 self.suite,
                 cx,
                 &client_hello.session_id,
-                chosen_share,
+                chosen_share_and_kxg,
                 chosen_psk_index,
                 resumedata
                     .as_ref()
@@ -346,14 +375,12 @@ mod client_hello {
                 emit_fake_ccs(cx.common);
             }
 
-            let (mut ocsp_response, mut sct_list) =
-                (server_key.get_ocsp(), server_key.get_sct_list());
+            let mut ocsp_response = server_key.get_ocsp();
             let doing_early_data = emit_encrypted_extensions(
                 &mut self.transcript,
                 self.suite,
                 cx,
                 &mut ocsp_response,
-                &mut sct_list,
                 client_hello,
                 resumedata.as_ref(),
                 self.extra_exts,
@@ -368,7 +395,6 @@ mod client_hello {
                     cx.common,
                     server_key.get_cert(),
                     ocsp_response,
-                    sct_list,
                 );
                 emit_certificate_verify_tls13(
                     &mut self.transcript,
@@ -456,19 +482,21 @@ mod client_hello {
         suite: &'static Tls13CipherSuite,
         cx: &mut ServerContext<'_>,
         session_id: &SessionId,
-        share: &KeyShareEntry,
+        share_and_kxgroup: (&KeyShareEntry, &'static dyn SupportedKxGroup),
         chosen_psk_idx: Option<usize>,
         resuming_psk: Option<&[u8]>,
         config: &ServerConfig,
     ) -> Result<KeyScheduleHandshake, Error> {
         let mut extensions = Vec::new();
 
-        // Prepare key exchange
-        let kx = kx::KeyExchange::choose(share.group, &config.kx_groups)
-            .and_then(kx::KeyExchange::start)
-            .ok_or(Error::FailedToGetRandomBytes)?;
+        // Prepare key exchange; the caller already found the matching SupportedKxGroup
+        let (share, kxgroup) = share_and_kxgroup;
+        debug_assert_eq!(kxgroup.name(), share.group);
+        let kx = kxgroup
+            .start()
+            .map_err(|_| Error::FailedToGetRandomBytes)?;
 
-        let kse = KeyShareEntry::new(share.group, kx.pubkey.as_ref());
+        let kse = KeyShareEntry::new(share.group, kx.pub_key());
         extensions.push(ServerExtension::KeyShare(kse));
         extensions.push(ServerExtension::SupportedVersions(ProtocolVersion::TLSv1_3));
 
@@ -515,9 +543,7 @@ mod client_hello {
         };
 
         // Do key exchange
-        let key_schedule = kx.complete(&share.payload.0, |secret| {
-            Ok(key_schedule_pre_handshake.into_handshake(secret))
-        })?;
+        let key_schedule = key_schedule_pre_handshake.into_handshake(kx, &share.payload.0)?;
 
         let handshake_hash = transcript.get_current_hash();
         let key_schedule = key_schedule.derive_server_handshake_secrets(
@@ -544,12 +570,13 @@ mod client_hello {
     fn emit_hello_retry_request(
         transcript: &mut HandshakeHash,
         suite: &'static Tls13CipherSuite,
+        session_id: SessionId,
         common: &mut CommonState,
         group: NamedGroup,
     ) {
         let mut req = HelloRetryRequest {
             legacy_version: ProtocolVersion::TLSv1_2,
-            session_id: SessionId::empty(),
+            session_id,
             cipher_suite: suite.common.suite,
             extensions: Vec::new(),
         };
@@ -625,7 +652,6 @@ mod client_hello {
         if early_data_configured && early_data_possible && !cx.data.early_data.was_rejected() {
             EarlyDataDecision::Accepted
         } else {
-            #[cfg(feature = "quic")]
             if cx.common.is_quic() {
                 // Clobber value set in tls13::emit_server_hello
                 cx.common.quic.early_secret = None;
@@ -640,22 +666,13 @@ mod client_hello {
         suite: &'static Tls13CipherSuite,
         cx: &mut ServerContext<'_>,
         ocsp_response: &mut Option<&[u8]>,
-        sct_list: &mut Option<&[u8]>,
         hello: &ClientHelloPayload,
         resumedata: Option<&persist::ServerSessionValue>,
         extra_exts: Vec<ServerExtension>,
         config: &ServerConfig,
     ) -> Result<EarlyDataDecision, Error> {
         let mut ep = hs::ExtensionProcessing::new();
-        ep.process_common(
-            config,
-            cx,
-            ocsp_response,
-            sct_list,
-            hello,
-            resumedata,
-            extra_exts,
-        )?;
+        ep.process_common(config, cx, ocsp_response, hello, resumedata, extra_exts)?;
 
         let early_data = decide_if_early_data_allowed(cx, hello, resumedata, suite, config);
         if early_data == EarlyDataDecision::Accepted {
@@ -685,7 +702,7 @@ mod client_hello {
             return Ok(false);
         }
 
-        let mut cr = CertificateRequestPayloadTLS13 {
+        let mut cr = CertificateRequestPayloadTls13 {
             context: PayloadU8::empty(),
             extensions: Vec::new(),
         };
@@ -696,21 +713,19 @@ mod client_hello {
         cr.extensions
             .push(CertReqExtension::SignatureAlgorithms(schemes.to_vec()));
 
-        let names = config
-            .verifier
-            .client_auth_root_subjects()
-            .to_vec();
-
-        if !names.is_empty() {
-            cr.extensions
-                .push(CertReqExtension::AuthorityNames(names));
-        }
+        cr.extensions
+            .push(CertReqExtension::AuthorityNames(
+                config
+                    .verifier
+                    .root_hint_subjects()
+                    .to_vec(),
+            ));
 
         let m = Message {
             version: ProtocolVersion::TLSv1_3,
             payload: MessagePayload::handshake(HandshakeMessagePayload {
                 typ: HandshakeType::CertificateRequest,
-                payload: HandshakePayload::CertificateRequestTLS13(cr),
+                payload: HandshakePayload::CertificateRequestTls13(cr),
             }),
         };
 
@@ -723,9 +738,8 @@ mod client_hello {
     fn emit_certificate_tls13(
         transcript: &mut HandshakeHash,
         common: &mut CommonState,
-        cert_chain: &[Certificate],
+        cert_chain: &[CertificateDer<'static>],
         ocsp_response: Option<&[u8]>,
-        sct_list: Option<&[u8]>,
     ) {
         let mut cert_entries = vec![];
         for cert in cert_chain {
@@ -746,21 +760,14 @@ mod client_hello {
                     .exts
                     .push(CertificateExtension::CertificateStatus(cst));
             }
-
-            // Likewise, SCT
-            if let Some(sct_list) = sct_list {
-                end_entity_cert
-                    .exts
-                    .push(CertificateExtension::make_sct(sct_list.to_owned()));
-            }
         }
 
-        let cert_body = CertificatePayloadTLS13::new(cert_entries);
+        let cert_body = CertificatePayloadTls13::new(cert_entries);
         let c = Message {
             version: ProtocolVersion::TLSv1_3,
             payload: MessagePayload::handshake(HandshakeMessagePayload {
                 typ: HandshakeType::Certificate,
-                payload: HandshakePayload::CertificateTLS13(cert_body),
+                payload: HandshakePayload::CertificateTls13(cert_body),
             }),
         };
 
@@ -775,12 +782,15 @@ mod client_hello {
         signing_key: &dyn sign::SigningKey,
         schemes: &[SignatureScheme],
     ) -> Result<(), Error> {
-        let message = verify::construct_tls13_server_verify_message(&transcript.get_current_hash());
+        let message = construct_server_verify_message(&transcript.get_current_hash());
 
         let signer = signing_key
             .choose_scheme(schemes)
             .ok_or_else(|| {
-                hs::incompatible(common, PeerIncompatible::NoSignatureSchemesInCommon)
+                common.send_fatal_alert(
+                    AlertDescription::HandshakeFailure,
+                    PeerIncompatible::NoSignatureSchemesInCommon,
+                )
             })?;
 
         let scheme = signer.scheme();
@@ -869,12 +879,12 @@ struct ExpectCertificate {
 
 impl State<ServerConnectionData> for ExpectCertificate {
     fn handle(mut self: Box<Self>, cx: &mut ServerContext<'_>, m: Message) -> hs::NextStateOrError {
-        let certp = require_handshake_msg!(
+        self.transcript.add_message(&m);
+        let certp = require_handshake_msg_move!(
             m,
             HandshakeType::Certificate,
-            HandshakePayload::CertificateTLS13
+            HandshakePayload::CertificateTls13
         )?;
-        self.transcript.add_message(&m);
 
         // We don't send any CertificateRequest extensions, so any extensions
         // here are illegal.
@@ -903,17 +913,17 @@ impl State<ServerConnectionData> for ExpectCertificate {
                     }));
                 }
 
-                cx.common
-                    .send_fatal_alert(AlertDescription::CertificateRequired);
-                return Err(Error::NoCertificatesPresented);
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::CertificateRequired,
+                    Error::NoCertificatesPresented,
+                ));
             }
             Some(chain) => chain,
         };
 
-        let now = std::time::SystemTime::now();
         self.config
             .verifier
-            .verify_client_cert(end_entity, intermediates, now)
+            .verify_client_cert(end_entity, intermediates, UnixTime::now())
             .map_err(|err| {
                 cx.common
                     .send_cert_verify_error_alert(err)
@@ -935,7 +945,7 @@ struct ExpectCertificateVerify {
     transcript: HandshakeHash,
     suite: &'static Tls13CipherSuite,
     key_schedule: KeyScheduleTrafficWithClientFinishedPending,
-    client_cert: Vec<Certificate>,
+    client_cert: CertificateChain,
     send_tickets: usize,
 }
 
@@ -950,7 +960,7 @@ impl State<ServerConnectionData> for ExpectCertificateVerify {
             let handshake_hash = self.transcript.get_current_hash();
             self.transcript.abandon_client_auth();
             let certs = &self.client_cert;
-            let msg = verify::construct_tls13_client_verify_message(&handshake_hash);
+            let msg = construct_client_verify_message(&handshake_hash);
 
             self.config
                 .verifier
@@ -998,11 +1008,10 @@ impl State<ServerConnectionData> for ExpectEarlyData {
                     .take_received_plaintext(payload)
                 {
                     true => Ok(self),
-                    false => {
-                        cx.common
-                            .send_fatal_alert(AlertDescription::UnexpectedMessage);
-                        Err(PeerMisbehaved::TooMuchEarlyDataReceived.into())
-                    }
+                    false => Err(cx.common.send_fatal_alert(
+                        AlertDescription::UnexpectedMessage,
+                        PeerMisbehaved::TooMuchEarlyDataReceived,
+                    )),
                 }
             }
             MessagePayload::Handshake {
@@ -1040,7 +1049,7 @@ fn get_server_session_value(
     key_schedule: &KeyScheduleTraffic,
     cx: &ServerContext<'_>,
     nonce: &[u8],
-    time_now: ticketer::TimeBase,
+    time_now: UnixTime,
     age_obfuscation_offset: u32,
 ) -> persist::ServerSessionValue {
     let version = ProtocolVersion::TLSv1_3;
@@ -1053,7 +1062,7 @@ fn get_server_session_value(
         cx.data.sni.as_ref(),
         version,
         suite.common.suite,
-        secret,
+        secret.as_ref(),
         cx.common.peer_certificates.clone(),
         cx.common.alpn_protocol.clone(),
         cx.data.resumption_data.clone(),
@@ -1078,12 +1087,19 @@ impl ExpectFinished {
         key_schedule: &KeyScheduleTraffic,
         config: &ServerConfig,
     ) -> Result<(), Error> {
-        let nonce = rand::random_vec(32)?;
-        let now = ticketer::TimeBase::now()?;
-        let age_add = rand::random_u32()?;
-        let plain =
-            get_server_session_value(transcript, suite, key_schedule, cx, &nonce, now, age_add)
-                .get_encoding();
+        let secure_random = config.provider.secure_random;
+        let nonce = rand::random_vec(secure_random, 32)?;
+        let age_add = rand::random_u32(secure_random)?;
+        let plain = get_server_session_value(
+            transcript,
+            suite,
+            key_schedule,
+            cx,
+            &nonce,
+            UnixTime::now(),
+            age_add,
+        )
+        .get_encoding();
 
         let stateless = config.ticketer.enabled();
         let (ticket, lifetime) = if stateless {
@@ -1093,7 +1109,7 @@ impl ExpectFinished {
             };
             (ticket, config.ticketer.lifetime())
         } else {
-            let id = rand::random_vec(32)?;
+            let id = rand::random_vec(secure_random, 32)?;
             let stored = config
                 .session_storage
                 .put(id.clone(), plain);
@@ -1105,7 +1121,7 @@ impl ExpectFinished {
             (id, stateful_lifetime)
         };
 
-        let mut payload = NewSessionTicketPayloadTLS13::new(lifetime, age_add, nonce, ticket);
+        let mut payload = NewSessionTicketPayloadTls13::new(lifetime, age_add, nonce, ticket);
 
         if config.max_early_data_size > 0 {
             if !stateless {
@@ -1125,7 +1141,7 @@ impl ExpectFinished {
             version: ProtocolVersion::TLSv1_3,
             payload: MessagePayload::handshake(HandshakeMessagePayload {
                 typ: HandshakeType::NewSessionTicket,
-                payload: HandshakePayload::NewSessionTicketTLS13(payload),
+                payload: HandshakePayload::NewSessionTicketTls13(payload),
             }),
         };
 
@@ -1145,16 +1161,16 @@ impl State<ServerConnectionData> for ExpectFinished {
             .key_schedule
             .sign_client_finish(&handshake_hash, cx.common);
 
-        let fin = constant_time::verify_slices_are_equal(expect_verify_data.as_ref(), &finished.0)
-            .map_err(|_| {
-                cx.common
-                    .send_fatal_alert(AlertDescription::DecryptError);
-                warn!("Finished wrong");
-                Error::DecryptError
-            })
-            .map(|_| verify::FinishedMessageVerified::assertion())?;
+        let fin = match ConstantTimeEq::ct_eq(expect_verify_data.as_ref(), &finished.0[..]).into() {
+            true => verify::FinishedMessageVerified::assertion(),
+            false => {
+                return Err(cx
+                    .common
+                    .send_fatal_alert(AlertDescription::DecryptError, Error::DecryptError));
+            }
+        };
 
-        // nb. future derivations include Client Finished, but not the
+        // Note: future derivations include Client Finished, but not the
         // main application data keying.
         self.transcript.add_message(&m);
 
@@ -1173,20 +1189,16 @@ impl State<ServerConnectionData> for ExpectFinished {
         // Application data may now flow, even if we have client auth enabled.
         cx.common.start_traffic();
 
-        #[cfg(feature = "quic")]
-        {
-            if cx.common.protocol == Protocol::Quic {
-                return Ok(Box::new(ExpectQuicTraffic {
-                    key_schedule: key_schedule_traffic,
-                    _fin_verified: fin,
-                }));
-            }
-        }
-
-        Ok(Box::new(ExpectTraffic {
-            key_schedule: key_schedule_traffic,
-            _fin_verified: fin,
-        }))
+        Ok(match cx.common.is_quic() {
+            true => Box::new(ExpectQuicTraffic {
+                key_schedule: key_schedule_traffic,
+                _fin_verified: fin,
+            }),
+            false => Box::new(ExpectTraffic {
+                key_schedule: key_schedule_traffic,
+                _fin_verified: fin,
+            }),
+        })
     }
 }
 
@@ -1202,13 +1214,11 @@ impl ExpectTraffic {
         common: &mut CommonState,
         key_update_request: &KeyUpdateRequest,
     ) -> Result<(), Error> {
-        #[cfg(feature = "quic")]
-        {
-            if let Protocol::Quic = common.protocol {
-                common.send_fatal_alert(AlertDescription::UnexpectedMessage);
-                warn!("KeyUpdate received in QUIC connection");
-                return Err(PeerMisbehaved::KeyUpdateReceivedInQuicConnection.into());
-            }
+        if let Protocol::Quic = common.protocol {
+            return Err(common.send_fatal_alert(
+                AlertDescription::UnexpectedMessage,
+                PeerMisbehaved::KeyUpdateReceivedInQuicConnection,
+            ));
         }
 
         common.check_aligned_handshake()?;
@@ -1261,20 +1271,17 @@ impl State<ServerConnectionData> for ExpectTraffic {
             .export_keying_material(output, label, context)
     }
 
-    #[cfg(feature = "secret_extraction")]
     fn extract_secrets(&self) -> Result<PartiallyExtractedSecrets, Error> {
         self.key_schedule
             .extract_secrets(Side::Server)
     }
 }
 
-#[cfg(feature = "quic")]
 struct ExpectQuicTraffic {
     key_schedule: KeyScheduleTraffic,
     _fin_verified: verify::FinishedMessageVerified,
 }
 
-#[cfg(feature = "quic")]
 impl State<ServerConnectionData> for ExpectQuicTraffic {
     fn handle(self: Box<Self>, _cx: &mut ServerContext<'_>, m: Message) -> hs::NextStateOrError {
         // reject all messages
