@@ -1,22 +1,16 @@
-use crate::cipher::{Iv, IvLen, MessageDecrypter};
 use crate::common_state::{CommonState, Side};
+use crate::crypto::cipher::{AeadKey, Iv, MessageDecrypter};
+use crate::crypto::tls13::{expand, Hkdf, HkdfExpander, OkmBlock, OutputLengthError};
+use crate::crypto::{hash, hmac, ActiveKeyExchange};
 use crate::error::Error;
-use crate::msgs::base::PayloadU8;
-#[cfg(feature = "quic")]
 use crate::quic;
-#[cfg(feature = "secret_extraction")]
-use crate::suites::{ConnectionTrafficSecrets, PartiallyExtractedSecrets};
+use crate::suites::PartiallyExtractedSecrets;
 use crate::{KeyLog, Tls13CipherSuite};
 
-/// Key schedule maintenance for TLS1.3
-use ring::{
-    aead,
-    digest::{self, Digest},
-    hkdf::{self, KeyType as _},
-    hmac,
-};
+use alloc::boxed::Box;
+use alloc::string::ToString;
 
-use super::{Tls13MessageDecrypter, Tls13MessageEncrypter};
+/// Key schedule maintenance for TLS1.3
 
 /// The kinds of secret we can extract from `KeySchedule`.
 #[derive(Debug, Clone, Copy, PartialEq)]
@@ -68,7 +62,7 @@ impl SecretKind {
 /// the type of hash.  This isn't used directly; but only through the
 /// typestates.
 struct KeySchedule {
-    current: hkdf::Prk,
+    current: Box<dyn HkdfExpander>,
     suite: &'static Tls13CipherSuite,
 }
 
@@ -92,7 +86,7 @@ impl KeyScheduleEarly {
 
     pub(crate) fn client_early_traffic_secret(
         &self,
-        hs_hash: &Digest,
+        hs_hash: &hash::Output,
         key_log: &dyn KeyLog,
         client_random: &[u8; 32],
         common: &mut CommonState,
@@ -113,7 +107,6 @@ impl KeyScheduleEarly {
                 .set_decrypter(&client_early_traffic_secret, common),
         }
 
-        #[cfg(feature = "quic")]
         if common.is_quic() {
             // If 0-RTT should be rejected, this will be clobbered by ExtensionProcessing
             // before the application can see.
@@ -123,7 +116,7 @@ impl KeyScheduleEarly {
 
     pub(crate) fn resumption_psk_binder_key_and_sign_verify_data(
         &self,
-        hs_hash: &Digest,
+        hs_hash: &hash::Output,
     ) -> hmac::Tag {
         let resumption_psk_binder_key = self
             .ks
@@ -148,9 +141,14 @@ impl KeySchedulePreHandshake {
         }
     }
 
-    pub(crate) fn into_handshake(mut self, secret: &[u8]) -> KeyScheduleHandshakeStart {
-        self.ks.input_secret(secret);
-        KeyScheduleHandshakeStart { ks: self.ks }
+    pub(crate) fn into_handshake(
+        mut self,
+        kx: Box<dyn ActiveKeyExchange>,
+        peer_public_key: &[u8],
+    ) -> Result<KeyScheduleHandshakeStart, Error> {
+        self.ks
+            .input_from_key_exchange(kx, peer_public_key)?;
+        Ok(KeyScheduleHandshakeStart { ks: self.ks })
     }
 }
 
@@ -169,7 +167,7 @@ impl KeyScheduleHandshakeStart {
     pub(crate) fn derive_client_handshake_secrets(
         mut self,
         early_data_enabled: bool,
-        hs_hash: Digest,
+        hs_hash: hash::Output,
         suite: &'static Tls13CipherSuite,
         key_log: &dyn KeyLog,
         client_random: &[u8; 32],
@@ -195,7 +193,7 @@ impl KeyScheduleHandshakeStart {
 
     pub(crate) fn derive_server_handshake_secrets(
         self,
-        hs_hash: Digest,
+        hs_hash: hash::Output,
         key_log: &dyn KeyLog,
         client_random: &[u8; 32],
         common: &mut CommonState,
@@ -213,10 +211,10 @@ impl KeyScheduleHandshakeStart {
 
     fn into_handshake(
         self,
-        hs_hash: Digest,
+        hs_hash: hash::Output,
         key_log: &dyn KeyLog,
         client_random: &[u8; 32],
-        _common: &mut CommonState,
+        common: &mut CommonState,
     ) -> KeyScheduleHandshake {
         // Use an empty handshake hash for the initial handshake.
         let client_secret = self.ks.derive_logged_secret(
@@ -233,14 +231,14 @@ impl KeyScheduleHandshakeStart {
             client_random,
         );
 
-        #[cfg(feature = "quic")]
-        if _common.is_quic() {
-            _common.quic.hs_secrets = Some(quic::Secrets::new(
+        if common.is_quic() {
+            common.quic.hs_secrets = Some(quic::Secrets::new(
                 client_secret.clone(),
                 server_secret.clone(),
                 self.ks.suite,
-                _common.side,
-                _common.quic.version,
+                self.ks.suite.quic.unwrap(),
+                common.side,
+                common.quic.version,
             ));
         }
 
@@ -254,12 +252,12 @@ impl KeyScheduleHandshakeStart {
 
 pub(crate) struct KeyScheduleHandshake {
     ks: KeySchedule,
-    client_handshake_traffic_secret: hkdf::Prk,
-    server_handshake_traffic_secret: hkdf::Prk,
+    client_handshake_traffic_secret: OkmBlock,
+    server_handshake_traffic_secret: OkmBlock,
 }
 
 impl KeyScheduleHandshake {
-    pub(crate) fn sign_server_finish(&self, hs_hash: &Digest) -> hmac::Tag {
+    pub(crate) fn sign_server_finish(&self, hs_hash: &hash::Output) -> hmac::Tag {
         self.ks
             .sign_finish(&self.server_handshake_traffic_secret, hs_hash)
     }
@@ -291,7 +289,7 @@ impl KeyScheduleHandshake {
 
     pub(crate) fn into_traffic_with_client_finished_pending(
         self,
-        hs_hash: Digest,
+        hs_hash: hash::Output,
         key_log: &dyn KeyLog,
         client_random: &[u8; 32],
         common: &mut CommonState,
@@ -308,12 +306,12 @@ impl KeyScheduleHandshake {
             .ks
             .set_encrypter(server_secret, common);
 
-        #[cfg(feature = "quic")]
         if common.is_quic() {
             common.quic.traffic_secrets = Some(quic::Secrets::new(
                 _client_secret.clone(),
                 server_secret.clone(),
                 traffic.ks.suite,
+                traffic.ks.suite.quic.unwrap(),
                 common.side,
                 common.quic.version,
             ));
@@ -327,8 +325,8 @@ impl KeyScheduleHandshake {
 
     pub(crate) fn into_pre_finished_client_traffic(
         self,
-        pre_finished_hash: Digest,
-        handshake_hash: Digest,
+        pre_finished_hash: hash::Output,
+        handshake_hash: hash::Output,
         key_log: &dyn KeyLog,
         client_random: &[u8; 32],
     ) -> (KeyScheduleClientBeforeFinished, hmac::Tag) {
@@ -363,12 +361,12 @@ impl KeyScheduleClientBeforeFinished {
             .ks
             .set_encrypter(client_secret, common);
 
-        #[cfg(feature = "quic")]
         if common.is_quic() {
             common.quic.traffic_secrets = Some(quic::Secrets::new(
                 client_secret.clone(),
                 server_secret.clone(),
                 self.traffic.ks.suite,
+                self.traffic.ks.suite.quic.unwrap(),
                 common.side,
                 common.quic.version,
             ));
@@ -382,7 +380,7 @@ impl KeyScheduleClientBeforeFinished {
 /// finished verify_data. The traffic stage key schedule can be extracted from it
 /// through signing the client finished hash.
 pub(crate) struct KeyScheduleTrafficWithClientFinishedPending {
-    handshake_client_traffic_secret: hkdf::Prk,
+    handshake_client_traffic_secret: OkmBlock,
     traffic: KeyScheduleTraffic,
 }
 
@@ -396,7 +394,7 @@ impl KeyScheduleTrafficWithClientFinishedPending {
 
     pub(crate) fn sign_client_finish(
         self,
-        hs_hash: &Digest,
+        hs_hash: &hash::Output,
         common: &mut CommonState,
     ) -> (KeyScheduleTraffic, hmac::Tag) {
         debug_assert_eq!(common.side, Side::Server);
@@ -421,15 +419,15 @@ impl KeyScheduleTrafficWithClientFinishedPending {
 /// to be available.
 pub(crate) struct KeyScheduleTraffic {
     ks: KeySchedule,
-    current_client_traffic_secret: hkdf::Prk,
-    current_server_traffic_secret: hkdf::Prk,
-    current_exporter_secret: hkdf::Prk,
+    current_client_traffic_secret: OkmBlock,
+    current_server_traffic_secret: OkmBlock,
+    current_exporter_secret: OkmBlock,
 }
 
 impl KeyScheduleTraffic {
     fn new(
         mut ks: KeySchedule,
-        hs_hash: Digest,
+        hs_hash: hash::Output,
         key_log: &dyn KeyLog,
         client_random: &[u8; 32],
     ) -> Self {
@@ -475,7 +473,7 @@ impl KeyScheduleTraffic {
         self.ks.set_decrypter(&secret, common);
     }
 
-    pub(crate) fn next_application_traffic_secret(&mut self, side: Side) -> hkdf::Prk {
+    pub(crate) fn next_application_traffic_secret(&mut self, side: Side) -> OkmBlock {
         let current = match side {
             Side::Client => &mut self.current_client_traffic_secret,
             Side::Server => &mut self.current_server_traffic_secret,
@@ -488,14 +486,12 @@ impl KeyScheduleTraffic {
 
     pub(crate) fn resumption_master_secret_and_derive_ticket_psk(
         &self,
-        hs_hash: &Digest,
+        hs_hash: &hash::Output,
         nonce: &[u8],
-    ) -> Vec<u8> {
-        let resumption_master_secret = self.ks.derive(
-            self.ks.algorithm(),
-            SecretKind::ResumptionMasterSecret,
-            hs_hash.as_ref(),
-        );
+    ) -> OkmBlock {
+        let resumption_master_secret = self
+            .ks
+            .derive(SecretKind::ResumptionMasterSecret, hs_hash.as_ref());
         self.ks
             .derive_ticket_psk(&resumption_master_secret, nonce)
     }
@@ -510,75 +506,40 @@ impl KeyScheduleTraffic {
             .export_keying_material(&self.current_exporter_secret, out, label, context)
     }
 
-    #[cfg(feature = "secret_extraction")]
     pub(crate) fn extract_secrets(&self, side: Side) -> Result<PartiallyExtractedSecrets, Error> {
-        fn expand<const KEY_LEN: usize, const IV_LEN: usize>(
-            secret: &hkdf::Prk,
-        ) -> Result<([u8; KEY_LEN], [u8; IV_LEN]), Error> {
-            let mut key = [0u8; KEY_LEN];
-            let mut iv = [0u8; IV_LEN];
+        fn expand(
+            secret: &OkmBlock,
+            hkdf: &'static dyn Hkdf,
+            aead_key_len: usize,
+        ) -> (AeadKey, Iv) {
+            let expander = hkdf.expander_for_okm(secret);
 
-            hkdf_expand_info(secret, PayloadU8Len(key.len()), b"key", &[], |okm| {
-                okm.fill(&mut key)
-            })
-            .map_err(|_| Error::General("hkdf_expand_info failed".to_string()))?;
-
-            hkdf_expand_info(secret, PayloadU8Len(iv.len()), b"iv", &[], |okm| {
-                okm.fill(&mut iv)
-            })
-            .map_err(|_| Error::General("hkdf_expand_info failed".to_string()))?;
-
-            Ok((key, iv))
+            (
+                hkdf_expand_label_aead_key(expander.as_ref(), aead_key_len, b"key", &[]),
+                hkdf_expand_label(expander.as_ref(), b"iv", &[]),
+            )
         }
 
-        let client_secrets;
-        let server_secrets;
-
-        let algo = self.ks.suite.common.aead_algorithm;
-        if algo == &ring::aead::AES_128_GCM {
-            let extract = |secret: &hkdf::Prk| -> Result<ConnectionTrafficSecrets, Error> {
-                let (key, iv_in) = expand::<16, 12>(secret)?;
-
-                let mut salt = [0u8; 4];
-                salt.copy_from_slice(&iv_in[..4]);
-
-                let mut iv = [0u8; 8];
-                iv.copy_from_slice(&iv_in[4..]);
-
-                Ok(ConnectionTrafficSecrets::Aes128Gcm { key, salt, iv })
-            };
-
-            client_secrets = extract(&self.current_client_traffic_secret)?;
-            server_secrets = extract(&self.current_server_traffic_secret)?;
-        } else if algo == &ring::aead::AES_256_GCM {
-            let extract = |secret: &hkdf::Prk| -> Result<ConnectionTrafficSecrets, Error> {
-                let (key, iv_in) = expand::<32, 12>(secret)?;
-
-                let mut salt = [0u8; 4];
-                salt.copy_from_slice(&iv_in[..4]);
-
-                let mut iv = [0u8; 8];
-                iv.copy_from_slice(&iv_in[4..]);
-
-                Ok(ConnectionTrafficSecrets::Aes256Gcm { key, salt, iv })
-            };
-
-            client_secrets = extract(&self.current_client_traffic_secret)?;
-            server_secrets = extract(&self.current_server_traffic_secret)?;
-        } else if algo == &ring::aead::CHACHA20_POLY1305 {
-            let extract = |secret: &hkdf::Prk| -> Result<ConnectionTrafficSecrets, Error> {
-                let (key, iv) = expand::<32, 12>(secret)?;
-                Ok(ConnectionTrafficSecrets::Chacha20Poly1305 { key, iv })
-            };
-
-            client_secrets = extract(&self.current_client_traffic_secret)?;
-            server_secrets = extract(&self.current_server_traffic_secret)?;
-        } else {
-            return Err(Error::General(format!(
-                "exporting secrets for {:?}: unimplemented",
-                algo
-            )));
-        }
+        let (client_key, client_iv) = expand(
+            &self.current_client_traffic_secret,
+            self.ks.suite.hkdf_provider,
+            self.ks.suite.aead_alg.key_len(),
+        );
+        let (server_key, server_iv) = expand(
+            &self.current_server_traffic_secret,
+            self.ks.suite.hkdf_provider,
+            self.ks.suite.aead_alg.key_len(),
+        );
+        let client_secrets = self
+            .ks
+            .suite
+            .aead_alg
+            .extract_keys(client_key, client_iv)?;
+        let server_secrets = self
+            .ks
+            .suite
+            .aead_alg
+            .extract_keys(server_key, server_iv)?;
 
         let (tx, rx) = match side {
             Side::Client => (client_secrets, server_secrets),
@@ -590,70 +551,88 @@ impl KeyScheduleTraffic {
 
 impl KeySchedule {
     fn new(suite: &'static Tls13CipherSuite, secret: &[u8]) -> Self {
-        let zeroes = [0u8; digest::MAX_OUTPUT_LEN];
-        let salt = hkdf::Salt::new(suite.hkdf_algorithm, &zeroes[..suite.hkdf_algorithm.len()]);
         Self {
-            current: salt.extract(secret),
+            current: suite
+                .hkdf_provider
+                .extract_from_secret(None, secret),
             suite,
         }
     }
 
-    fn set_encrypter(&self, secret: &hkdf::Prk, common: &mut CommonState) {
-        let key = derive_traffic_key(secret, self.suite.common.aead_algorithm);
-        let iv = derive_traffic_iv(secret);
+    fn set_encrypter(&self, secret: &OkmBlock, common: &mut CommonState) {
+        let expander = self
+            .suite
+            .hkdf_provider
+            .expander_for_okm(secret);
+        let key = derive_traffic_key(expander.as_ref(), self.suite.aead_alg.key_len());
+        let iv = derive_traffic_iv(expander.as_ref());
 
         common
             .record_layer
-            .set_message_encrypter(Box::new(Tls13MessageEncrypter {
-                enc_key: aead::LessSafeKey::new(key),
-                iv,
-            }));
+            .set_message_encrypter(self.suite.aead_alg.encrypter(key, iv));
     }
 
-    fn set_decrypter(&self, secret: &hkdf::Prk, common: &mut CommonState) {
+    fn set_decrypter(&self, secret: &OkmBlock, common: &mut CommonState) {
         common
             .record_layer
             .set_message_decrypter(self.derive_decrypter(secret));
     }
 
-    fn derive_decrypter(&self, secret: &hkdf::Prk) -> Box<dyn MessageDecrypter> {
-        let key = derive_traffic_key(secret, self.suite.common.aead_algorithm);
-        let iv = derive_traffic_iv(secret);
-        Box::new(Tls13MessageDecrypter {
-            dec_key: aead::LessSafeKey::new(key),
-            iv,
-        })
-    }
-
-    #[inline]
-    fn algorithm(&self) -> hkdf::Algorithm {
-        self.suite.hkdf_algorithm
+    fn derive_decrypter(&self, secret: &OkmBlock) -> Box<dyn MessageDecrypter> {
+        let expander = self
+            .suite
+            .hkdf_provider
+            .expander_for_okm(secret);
+        let key = derive_traffic_key(expander.as_ref(), self.suite.aead_alg.key_len());
+        let iv = derive_traffic_iv(expander.as_ref());
+        self.suite.aead_alg.decrypter(key, iv)
     }
 
     fn new_with_empty_secret(suite: &'static Tls13CipherSuite) -> Self {
-        let zeroes = [0u8; digest::MAX_OUTPUT_LEN];
-        Self::new(suite, &zeroes[..suite.hkdf_algorithm.len()])
+        Self {
+            current: suite
+                .hkdf_provider
+                .extract_from_zero_ikm(None),
+            suite,
+        }
     }
 
     /// Input the empty secret.
     fn input_empty(&mut self) {
-        let zeroes = [0u8; digest::MAX_OUTPUT_LEN];
-        self.input_secret(&zeroes[..self.suite.hkdf_algorithm.len()]);
+        let salt = self.derive_for_empty_hash(SecretKind::DerivedSecret);
+        self.current = self
+            .suite
+            .hkdf_provider
+            .extract_from_zero_ikm(Some(salt.as_ref()));
     }
 
     /// Input the given secret.
+    #[cfg(all(test, any(feature = "ring", feature = "aws_lc_rs")))]
     fn input_secret(&mut self, secret: &[u8]) {
-        let salt: hkdf::Salt = self.derive_for_empty_hash(SecretKind::DerivedSecret);
-        self.current = salt.extract(secret);
+        let salt = self.derive_for_empty_hash(SecretKind::DerivedSecret);
+        self.current = self
+            .suite
+            .hkdf_provider
+            .extract_from_secret(Some(salt.as_ref()), secret);
+    }
+
+    /// Input the shared secret resulting from completing the given key exchange.
+    fn input_from_key_exchange(
+        &mut self,
+        kx: Box<dyn ActiveKeyExchange>,
+        peer_public_key: &[u8],
+    ) -> Result<(), Error> {
+        let salt = self.derive_for_empty_hash(SecretKind::DerivedSecret);
+        self.current = self
+            .suite
+            .hkdf_provider
+            .extract_from_kx_shared_secret(Some(salt.as_ref()), kx, peer_public_key)?;
+        Ok(())
     }
 
     /// Derive a secret of given `kind`, using current handshake hash `hs_hash`.
-    fn derive<T, L>(&self, key_type: L, kind: SecretKind, hs_hash: &[u8]) -> T
-    where
-        T: for<'a> From<hkdf::Okm<'a, L>>,
-        L: hkdf::KeyType,
-    {
-        hkdf_expand(&self.current, key_type, kind.to_bytes(), hs_hash)
+    fn derive(&self, kind: SecretKind, hs_hash: &[u8]) -> OkmBlock {
+        hkdf_expand_label_block(self.current.as_ref(), kind.to_bytes(), hs_hash)
     }
 
     fn derive_logged_secret(
@@ -662,131 +641,178 @@ impl KeySchedule {
         hs_hash: &[u8],
         key_log: &dyn KeyLog,
         client_random: &[u8; 32],
-    ) -> hkdf::Prk {
+    ) -> OkmBlock {
+        let output = self.derive(kind, hs_hash);
+
         let log_label = kind
             .log_label()
             .expect("not a loggable secret");
         if key_log.will_log(log_label) {
-            let secret = self
-                .derive::<PayloadU8, _>(
-                    PayloadU8Len(self.suite.hkdf_algorithm.len()),
-                    kind,
-                    hs_hash,
-                )
-                .into_inner();
-            key_log.log(log_label, client_random, &secret);
+            key_log.log(log_label, client_random, output.as_ref());
         }
-        self.derive(self.suite.hkdf_algorithm, kind, hs_hash)
+        output
     }
 
     /// Derive a secret of given `kind` using the hash of the empty string
     /// for the handshake hash.  Useful only for
     /// `SecretKind::ResumptionPSKBinderKey` and
     /// `SecretKind::DerivedSecret`.
-    fn derive_for_empty_hash<T>(&self, kind: SecretKind) -> T
-    where
-        T: for<'a> From<hkdf::Okm<'a, hkdf::Algorithm>>,
-    {
-        let digest_alg = self
+    fn derive_for_empty_hash(&self, kind: SecretKind) -> OkmBlock {
+        let empty_hash = self
             .suite
-            .hkdf_algorithm
-            .hmac_algorithm()
-            .digest_algorithm();
-        let empty_hash = digest::digest(digest_alg, &[]);
-        self.derive(self.suite.hkdf_algorithm, kind, empty_hash.as_ref())
+            .common
+            .hash_provider
+            .start()
+            .finish();
+        self.derive(kind, empty_hash.as_ref())
     }
 
     /// Sign the finished message consisting of `hs_hash` using a current
     /// traffic secret.
-    fn sign_finish(&self, base_key: &hkdf::Prk, hs_hash: &Digest) -> hmac::Tag {
+    fn sign_finish(&self, base_key: &OkmBlock, hs_hash: &hash::Output) -> hmac::Tag {
         self.sign_verify_data(base_key, hs_hash)
     }
 
     /// Sign the finished message consisting of `hs_hash` using the key material
     /// `base_key`.
-    fn sign_verify_data(&self, base_key: &hkdf::Prk, hs_hash: &Digest) -> hmac::Tag {
-        let hmac_alg = self
+    fn sign_verify_data(&self, base_key: &OkmBlock, hs_hash: &hash::Output) -> hmac::Tag {
+        let expander = self
             .suite
-            .hkdf_algorithm
-            .hmac_algorithm();
-        let hmac_key = hkdf_expand(base_key, hmac_alg, b"finished", &[]);
-        hmac::sign(&hmac_key, hs_hash.as_ref())
+            .hkdf_provider
+            .expander_for_okm(base_key);
+        let hmac_key = hkdf_expand_label_block(expander.as_ref(), b"finished", &[]);
+
+        self.suite
+            .hkdf_provider
+            .hmac_sign(&hmac_key, hs_hash.as_ref())
     }
 
     /// Derive the next application traffic secret, returning it.
-    fn derive_next(&self, base_key: &hkdf::Prk) -> hkdf::Prk {
-        hkdf_expand(base_key, self.suite.hkdf_algorithm, b"traffic upd", &[])
+    fn derive_next(&self, base_key: &OkmBlock) -> OkmBlock {
+        let expander = self
+            .suite
+            .hkdf_provider
+            .expander_for_okm(base_key);
+        hkdf_expand_label_block(expander.as_ref(), b"traffic upd", &[])
     }
 
     /// Derive the PSK to use given a resumption_master_secret and
     /// ticket_nonce.
-    fn derive_ticket_psk(&self, rms: &hkdf::Prk, nonce: &[u8]) -> Vec<u8> {
-        let payload: PayloadU8 = hkdf_expand(
-            rms,
-            PayloadU8Len(self.suite.hkdf_algorithm.len()),
-            b"resumption",
-            nonce,
-        );
-        payload.into_inner()
+    fn derive_ticket_psk(&self, rms: &OkmBlock, nonce: &[u8]) -> OkmBlock {
+        let expander = self
+            .suite
+            .hkdf_provider
+            .expander_for_okm(rms);
+        hkdf_expand_label_block(expander.as_ref(), b"resumption", nonce)
     }
 
     fn export_keying_material(
         &self,
-        current_exporter_secret: &hkdf::Prk,
+        current_exporter_secret: &OkmBlock,
         out: &mut [u8],
         label: &[u8],
         context: Option<&[u8]>,
     ) -> Result<(), Error> {
-        let digest_alg = self
+        let secret = {
+            let h_empty = self
+                .suite
+                .common
+                .hash_provider
+                .hash(&[]);
+
+            let expander = self
+                .suite
+                .hkdf_provider
+                .expander_for_okm(current_exporter_secret);
+            hkdf_expand_label_block(expander.as_ref(), label, h_empty.as_ref())
+        };
+
+        let h_context = self
             .suite
-            .hkdf_algorithm
-            .hmac_algorithm()
-            .digest_algorithm();
+            .common
+            .hash_provider
+            .hash(context.unwrap_or(&[]));
 
-        let h_empty = digest::digest(digest_alg, &[]);
-        let secret: hkdf::Prk = hkdf_expand(
-            current_exporter_secret,
-            self.suite.hkdf_algorithm,
-            label,
-            h_empty.as_ref(),
-        );
-
-        let h_context = digest::digest(digest_alg, context.unwrap_or(&[]));
-
-        // TODO: Test what happens when this fails
-        hkdf_expand_info(
-            &secret,
-            PayloadU8Len(out.len()),
-            b"exporter",
-            h_context.as_ref(),
-            |okm| okm.fill(out),
-        )
-        .map_err(|_| Error::General("exporting too much".to_string()))
+        let expander = self
+            .suite
+            .hkdf_provider
+            .expander_for_okm(&secret);
+        hkdf_expand_label_slice(expander.as_ref(), b"exporter", h_context.as_ref(), out)
+            .map_err(|_| Error::General("exporting too much".to_string()))
     }
 }
 
-pub(crate) fn hkdf_expand<T, L>(secret: &hkdf::Prk, key_type: L, label: &[u8], context: &[u8]) -> T
-where
-    T: for<'a> From<hkdf::Okm<'a, L>>,
-    L: hkdf::KeyType,
-{
-    hkdf_expand_info(secret, key_type, label, context, |okm| okm.into())
-}
-
-fn hkdf_expand_info<F, T, L>(
-    secret: &hkdf::Prk,
-    key_type: L,
+/// [HKDF-Expand-Label] where the output length is a compile-time constant, and therefore
+/// it is infallible.
+///
+/// [HKDF-Expand-Label]: <https://www.rfc-editor.org/rfc/rfc8446#section-7.1>
+pub(crate) fn hkdf_expand_label<T: From<[u8; N]>, const N: usize>(
+    expander: &dyn HkdfExpander,
     label: &[u8],
     context: &[u8],
+) -> T {
+    hkdf_expand_label_inner(expander, label, context, N, |e, info| expand(e, info))
+}
+
+/// [HKDF-Expand-Label] where the output is one block in size.
+pub(crate) fn hkdf_expand_label_block(
+    expander: &dyn HkdfExpander,
+    label: &[u8],
+    context: &[u8],
+) -> OkmBlock {
+    hkdf_expand_label_inner(expander, label, context, expander.hash_len(), |e, info| {
+        e.expand_block(info)
+    })
+}
+
+/// [HKDF-Expand-Label] where the output is an AEAD key.
+pub(crate) fn hkdf_expand_label_aead_key(
+    expander: &dyn HkdfExpander,
+    key_len: usize,
+    label: &[u8],
+    context: &[u8],
+) -> AeadKey {
+    hkdf_expand_label_inner(expander, label, context, key_len, |e, info| {
+        let key: AeadKey = expand(e, info);
+        key.with_length(key_len)
+    })
+}
+
+/// [HKDF-Expand-Label] where the output is a slice.
+///
+/// This can fail because HKDF-Expand is limited in its maximum output length.
+fn hkdf_expand_label_slice(
+    expander: &dyn HkdfExpander,
+    label: &[u8],
+    context: &[u8],
+    output: &mut [u8],
+) -> Result<(), OutputLengthError> {
+    hkdf_expand_label_inner(expander, label, context, output.len(), |e, info| {
+        e.expand_slice(info, output)
+    })
+}
+
+pub(crate) fn derive_traffic_key(expander: &dyn HkdfExpander, aead_key_len: usize) -> AeadKey {
+    hkdf_expand_label_aead_key(expander, aead_key_len, b"key", &[])
+}
+
+pub(crate) fn derive_traffic_iv(expander: &dyn HkdfExpander) -> Iv {
+    hkdf_expand_label(expander, b"iv", &[])
+}
+
+fn hkdf_expand_label_inner<F, T>(
+    expander: &dyn HkdfExpander,
+    label: &[u8],
+    context: &[u8],
+    n: usize,
     f: F,
 ) -> T
 where
-    F: for<'b> FnOnce(hkdf::Okm<'b, L>) -> T,
-    L: hkdf::KeyType,
+    F: FnOnce(&dyn HkdfExpander, &[&[u8]]) -> T,
 {
     const LABEL_PREFIX: &[u8] = b"tls13 ";
 
-    let output_len = u16::to_be_bytes(key_type.len() as u16);
+    let output_len = u16::to_be_bytes(n as u16);
     let label_len = u8::to_be_bytes((LABEL_PREFIX.len() + label.len()) as u8);
     let context_len = u8::to_be_bytes(context.len() as u8);
 
@@ -798,43 +824,20 @@ where
         &context_len[..],
         context,
     ];
-    let okm = secret.expand(info, key_type).unwrap();
 
-    f(okm)
+    f(expander, info)
 }
 
-pub(crate) struct PayloadU8Len(pub(crate) usize);
-impl hkdf::KeyType for PayloadU8Len {
-    fn len(&self) -> usize {
-        self.0
-    }
-}
+#[cfg(all(test, any(feature = "ring", feature = "aws_lc_rs")))]
+mod tests {
+    use core::fmt::Debug;
 
-impl From<hkdf::Okm<'_, PayloadU8Len>> for PayloadU8 {
-    fn from(okm: hkdf::Okm<PayloadU8Len>) -> Self {
-        let mut r = vec![0u8; okm.len().0];
-        okm.fill(&mut r[..]).unwrap();
-        Self::new(r)
-    }
-}
-
-pub(crate) fn derive_traffic_key(
-    secret: &hkdf::Prk,
-    aead_algorithm: &'static aead::Algorithm,
-) -> aead::UnboundKey {
-    hkdf_expand(secret, aead_algorithm, b"key", &[])
-}
-
-pub(crate) fn derive_traffic_iv(secret: &hkdf::Prk) -> Iv {
-    hkdf_expand(secret, IvLen, b"iv", &[])
-}
-
-#[cfg(test)]
-mod test {
     use super::{derive_traffic_iv, derive_traffic_key, KeySchedule, SecretKind};
-    use crate::tls13::TLS13_CHACHA20_POLY1305_SHA256_INTERNAL;
+    use crate::test_provider::ring_like::aead;
+    use crate::test_provider::tls13::{
+        TLS13_AES_128_GCM_SHA256_INTERNAL, TLS13_CHACHA20_POLY1305_SHA256_INTERNAL,
+    };
     use crate::KeyLog;
-    use ring::aead;
 
     #[test]
     fn test_vectors() {
@@ -967,6 +970,7 @@ mod test {
         expected_key: &[u8],
         expected_iv: &[u8],
     ) {
+        #[derive(Debug)]
         struct Log<'a>(&'a [u8]);
         impl KeyLog for Log<'_> {
             fn log(&self, _label: &str, _client_random: &[u8], secret: &[u8]) {
@@ -978,15 +982,19 @@ mod test {
 
         // Since we can't test key equality, we test the output of sealing with the key instead.
         let aead_alg = &aead::AES_128_GCM;
-        let key = derive_traffic_key(&traffic_secret, aead_alg);
+        let expander = TLS13_AES_128_GCM_SHA256_INTERNAL
+            .hkdf_provider
+            .expander_for_okm(&traffic_secret);
+        let key = derive_traffic_key(expander.as_ref(), aead_alg.key_len());
+        let key = aead::UnboundKey::new(aead_alg, key.as_ref()).unwrap();
         let seal_output = seal_zeroes(key);
         let expected_key = aead::UnboundKey::new(aead_alg, expected_key).unwrap();
         let expected_seal_output = seal_zeroes(expected_key);
         assert_eq!(seal_output, expected_seal_output);
         assert!(seal_output.len() >= 48); // Sanity check.
 
-        let iv = derive_traffic_iv(&traffic_secret);
-        assert_eq!(iv.value(), expected_iv);
+        let iv = derive_traffic_iv(expander.as_ref());
+        assert_eq!(iv.as_ref(), expected_iv);
     }
 
     fn seal_zeroes(key: aead::UnboundKey) -> Vec<u8> {
@@ -1001,3 +1009,52 @@ mod test {
         seal_output
     }
 }
+
+#[cfg(bench)]
+mod benchmarks {
+    #[cfg(any(feature = "ring", feature = "aws_lc_rs"))]
+    #[bench]
+    fn bench_sha256(b: &mut test::Bencher) {
+        use core::fmt::Debug;
+
+        use super::{derive_traffic_iv, derive_traffic_key, KeySchedule, SecretKind};
+        use crate::test_provider::tls13::TLS13_CHACHA20_POLY1305_SHA256_INTERNAL;
+        use crate::KeyLog;
+
+        fn extract_traffic_secret(ks: &KeySchedule, kind: SecretKind) {
+            #[derive(Debug)]
+            struct Log;
+
+            impl KeyLog for Log {
+                fn log(&self, _label: &str, _client_random: &[u8], _secret: &[u8]) {}
+            }
+
+            let hash = [0u8; 32];
+            let traffic_secret = ks.derive_logged_secret(kind, &hash, &Log, &[0u8; 32]);
+            let traffic_secret_expander = TLS13_CHACHA20_POLY1305_SHA256_INTERNAL
+                .hkdf_provider
+                .expander_for_okm(&traffic_secret);
+            test::black_box(derive_traffic_key(
+                traffic_secret_expander.as_ref(),
+                TLS13_CHACHA20_POLY1305_SHA256_INTERNAL
+                    .aead_alg
+                    .key_len(),
+            ));
+            test::black_box(derive_traffic_iv(traffic_secret_expander.as_ref()));
+        }
+
+        b.iter(|| {
+            let mut ks =
+                KeySchedule::new_with_empty_secret(TLS13_CHACHA20_POLY1305_SHA256_INTERNAL);
+            ks.input_secret(&[0u8; 32]);
+
+            extract_traffic_secret(&ks, SecretKind::ClientHandshakeTrafficSecret);
+            extract_traffic_secret(&ks, SecretKind::ServerHandshakeTrafficSecret);
+
+            ks.input_empty();
+
+            extract_traffic_secret(&ks, SecretKind::ClientApplicationTrafficSecret);
+            extract_traffic_secret(&ks, SecretKind::ServerApplicationTrafficSecret);
+        });
+    }
+}
